from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource


class DatasetTemplate(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    DeletableAPIResource,
):
    """
    DatasetTemplates contain the schema of a Dataset, including some
    properties and all the fields.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("name", "Name"),
        ("description", "Description"),
    )

    @property
    def import_params(self):
        """
        Get DatasetImport parameters from a template
        and format them correctly.
        """
        return {
            "target_fields": self.fields,
            "reader_params": self.reader_params,
            "entity_params": self.entity_params,
            "annotator_params": self.annotator_params,
            "validation_params": self.validation_params,
        }



import asyncio
import os

import streamlit as st
import quartzbio

from quartzbio_auth import QuartzBioOAuth2


class QuartzBioStreamlit:
    """QuartzBio OAuth2 wrapper for restricting access to Streamlit apps"""

    # App settings loaded from environment variables or .env file
    CLIENT_ID = os.environ.get("CLIENT_ID", "Application (client) Id")
    CLIENT_SECRET = os.environ.get("CLIENT_SECRET", "Application (client) secret")
    APP_URL = os.environ.get("APP_URL", "http://localhost:5000")

    def quartzbio_login_component(self, authorization_url):
        """Streamlit component for logging into QuartzBio"""

        st.title("Secure Streamlit App")
        st.write(
            """
            <h4>
                <a target="_self" href="{}">Log in to QuartzBio to continue</a>
            </h4>
            This app requires a QuartzBio account. <br>
            <a href="mailto:support@quartzbio.com">Contact Support</a>
            """.format(authorization_url),
            unsafe_allow_html=True,
        )

    def get_token_from_session(self):
        """Reads token from streamlit session state.

        If token is in the session state then the user is authorized to use the app and vice versa.
        """

        if "token" not in st.session_state:
            token = None
        else:
            token = st.session_state.token

        return token

    def wrap(self, streamlit_app):
        """QuartzBio OAuth2 wrapper around streamlit app"""

        # QuartzBio OAuth2 client
        oauth_client = QuartzBioOAuth2(self.CLIENT_ID, self.CLIENT_SECRET)
        authorization_url = oauth_client.get_authorization_url(
            self.APP_URL,
        )

        # Authorization token from Streamlit session state
        oauth_token = self.get_token_from_session()

        if oauth_token is None:
            # User is not authrized to use the app
            try:
                # Trying to get the authorization token from the url if successfully authorized
                code = st.experimental_get_query_params()["code"]

                # Remove authorization token from the url params
                params = {}
                st.experimental_set_query_params(**params)

            except:
                # Display QuartzBio login until user is successfully authorized
                self.quartzbio_login_component(authorization_url)
            else:
                try:
                    # Getting the token from token API by sending the authorization code
                    oauth_token = asyncio.run(
                        oauth_client.get_access_token(code, self.APP_URL)
                    )
                except:
                    st.error(
                        "This account is not allowed or page was refreshed. Please login again."
                    )
                    self.quartzbio_login_component(authorization_url)
                else:
                    # Check if token has expired:
                    if oauth_token.is_expired():
                        st.error("Login session has ended. Please login again.")
                        self.quartzbio_login_component(authorization_url)
                    else:
                        # User is now authenticated and authorized to use the app

                        # QuartzBioClient to acces API
                        quartzbio_client = quartzbio.QuartzBioClient(
                            token=oauth_token["access_token"],
                            token_type=oauth_token["token_type"],
                        )

                        # Saving token and quartzbio client to the Streamlit session state
                        st.session_state.quartzbio_client = quartzbio_client
                        st.session_state.token = oauth_token

                        streamlit_app()
        else:
            # User is authorized to the the app
            streamlit_app()



# -*- coding: utf-8 -*-

from ..client import client
from .util import json


def convert_to_quartzbio_object(resp, **kwargs):
    from . import types

    _client = kwargs.pop("client", None)

    if isinstance(resp, list):
        return [convert_to_quartzbio_object(i, client=_client) for i in resp]
    elif isinstance(resp, dict) and not isinstance(resp, QuartzBioObject):
        resp = resp.copy()
        klass_name = resp.get("class_name")
        if isinstance(klass_name, str):
            klass = types.get(klass_name, QuartzBioObject)
        else:
            klass = QuartzBioObject
        return klass.construct_from(resp, client=_client)
    else:
        return resp


class QuartzBioObject(dict):
    """Base class for all QuartzBio API resource objects"""

    ID_ATTR = "id"

    # Allows pre-setting a QuartzBioClient
    _client = None

    def __init__(self, id=None, **params):
        super(QuartzBioObject, self).__init__()

        self._client = params.pop("client", self._client or client)

        # store manually updated values for partial updates
        self._unsaved_values = set()

        if id:
            self[self.ID_ATTR] = id

    def __setattr__(self, k, v):
        if k[0] == "_" or k in self.__dict__:
            return super(QuartzBioObject, self).__setattr__(k, v)
        else:
            self[k] = v

    def __getattr__(self, k):
        if k[0] == "_":
            raise AttributeError(k)

        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __setitem__(self, k, v):
        super(QuartzBioObject, self).__setitem__(k, v)
        self._unsaved_values.add(k)

    @classmethod
    def construct_from(cls, values, **kwargs):
        """Used to create a new object from an HTTP response"""
        instance = cls(values.get(cls.ID_ATTR), **kwargs)
        instance.refresh_from(values)
        return instance

    def refresh_from(self, values):
        self.clear()
        self._unsaved_values = set()

        for k, v in values.items():
            super(QuartzBioObject, self).__setitem__(
                k, convert_to_quartzbio_object(v, client=self._client)
            )

    def request(self, method, url, **kwargs):
        response = self._client.request(method, url, **kwargs)
        return convert_to_quartzbio_object(response, client=self._client)

    def __repr__(self):
        if isinstance(self.get("class_name"), str):
            ident_parts = [self.get("class_name")]
        else:
            ident_parts = [type(self).__name__]

        if isinstance(self.get(self.ID_ATTR), int):
            ident_parts.append(
                "%s=%d"
                % (
                    self.ID_ATTR,
                    self.get(self.ID_ATTR),
                )
            )

        _repr = "<%s at %s> JSON: %s" % (
            " ".join(ident_parts),
            hex(id(self)),
            str(self),
        )

        return _repr

    def __str__(self):
        return json.dumps(self, sort_keys=True, indent=2)

    @property
    def quartzbio_id(self):
        return self.id



from dash.dependencies import Input, Output
import dash
import dash_html_components as html
import dash_core_components as dcc
import time

from quartzbio.contrib.dash import QuartzBioAuth

from .IntegrationTests import IntegrationTests
from .utils import switch_windows

from .credentials import OAUTH_CLIENT_ID
from .credentials import OAUTH_USERNAME
from .credentials import OAUTH_PASSWORD
from .credentials import OAUTH_DOMAIN


class Tests(IntegrationTests):
    test_client_id = OAUTH_CLIENT_ID
    test_domain = OAUTH_DOMAIN
    test_username = OAUTH_USERNAME
    test_password = OAUTH_PASSWORD

    def quartzbio_auth_login_flow(self, username, pw, domain, url_base_pathname):
        app = dash.Dash(__name__, url_base_pathname=url_base_pathname)
        app.layout = html.Div(
            [dcc.Input(id="input", value="initial value"), html.Div(id="output")]
        )

        @app.callback(Output("output", "children"), [Input("input", "value")])
        def update_output(new_value):
            return new_value

        QuartzBioAuth(
            app,
            "http://localhost:8050{}".format(url_base_pathname),
            self.test_client_id,
        )

        self.startServer(app)

        time.sleep(10)
        self.percy_snapshot(
            "login screen - {} {} {}".format(username, pw, url_base_pathname)
        )
        try:
            self.wait_for_element_by_id("dash-auth--login__container")
        except Exception as e:
            print(self.wait_for_element_by_tag_name("body").html)
            raise e

        self.driver.find_element_by_id("dash-auth--login__button").click()
        time.sleep(5)
        switch_windows(self.driver)

        # Domain selection
        time.sleep(5)
        self.wait_for_element_by_css_selector('input[name="domain"]').send_keys(domain)
        self.driver.find_element_by_css_selector('button[type="submit"]').click()

        # Login page
        time.sleep(10)
        self.wait_for_element_by_css_selector('input[name="email"]').send_keys(username)
        self.wait_for_element_by_css_selector('input[name="password"]').send_keys(pw)
        self.driver.find_element_by_css_selector('button[type="submit"]').click()

        # OAuth authorize page
        time.sleep(10)
        self.percy_snapshot(
            "oauth screen - {} {} {}".format(username, pw, url_base_pathname)
        )
        self.wait_for_element_by_css_selector('button[type="submit"]').click()

    def private_app_authorized(self, url_base_pathname):
        self.quartzbio_auth_login_flow(
            self.test_username, self.test_password, self.test_domain, url_base_pathname
        )
        switch_windows(self.driver)
        time.sleep(5)
        self.percy_snapshot("private_app_authorized - {}".format(url_base_pathname))
        try:
            el = self.wait_for_element_by_id("output")
        except:
            print((self.driver.find_element_by_tag_name("body").html))
        self.assertEqual(el.text, "initial value")

    def test_private_app_authorized_index(self):
        self.private_app_authorized("/")



import os
import hashlib
import binascii

# Default thresholds for multipart S3 files
MULTIPART_THRESHOLD = 64 * 1024 * 1024
MULTIPART_CHUNKSIZE = 64 * 1024 * 1024


def md5sum(
    path,
    multipart_threshold=MULTIPART_THRESHOLD,
    multipart_chunksize=MULTIPART_CHUNKSIZE,
):
    def _read_chunks(f, chunk_size):
        chunk = f.read(chunk_size)
        while chunk:
            yield chunk
            chunk = f.read(chunk_size)

    filesize = os.path.getsize(path)

    with open(path, "rb") as f:
        if multipart_threshold and filesize > multipart_threshold:
            block_count = 0
            md5string = ""
            for block in _read_chunks(f, multipart_chunksize):
                md5 = hashlib.md5()
                md5.update(block)
                md5string += md5.hexdigest()
                block_count += 1

            md5 = hashlib.md5()
            md5.update(binascii.unhexlify(md5string))
        else:
            block_count = None
            md5 = hashlib.md5()
            for block in _read_chunks(f, multipart_chunksize):
                md5.update(block)

    return md5.hexdigest(), block_count



# -*- coding: utf-8 -*-

import json
import time
import inspect
import os
from typing import Literal

import quartzbio

from .version import VERSION
from .errors import QuartzBioError
from .auth import authenticate, QuartzBioTokenAuth

import platform
import requests
import textwrap
import logging

from requests import Session, codes, adapters
from urllib3.util.retry import Retry

from urllib.parse import urljoin

# Try using pyopenssl if available.
# Requires: pip install pyopenssl ndg-httpsclient pyasn1
# See http://urllib3.readthedocs.org/en/latest/contrib.html#module-urllib3.contrib.pyopenssl  # noqa
import ssl
import sys

# Python 3.8+ compatibility
try:
    import urllib3

    if sys.version_info <= (3, 9):
        import urllib3.contrib.pyopenssl

        urllib3.contrib.pyopenssl.inject_into_urllib3()
    else:
        # Python 3.10+ automatically handles SSL; no need for inject_into_urllib3()
        pass
except ImportError:
    pass

# Ensure SSL/TLS support is available
if not ssl.HAS_TLSv1_2:
    raise RuntimeError("TLS 1.2 support is required but not available.")


logger = logging.getLogger("quartzbio")


def _handle_api_error(response):
    if response.status_code not in [400, 401, 403, 404]:
        logger.info("API Error: %d" % response.status_code)
    raise QuartzBioError(response=response)


def _handle_request_error(e):
    if isinstance(e, requests.exceptions.RequestException):
        msg = QuartzBioError.default_message
        err = "%s: %s" % (type(e).__name__, str(e))
    else:
        msg = (
            "Unexpected error communicating with QuartzBio.\n"
            "It looks like there's probably a configuration "
            "issue locally.\nIf this problem persists, let us "
            "know at support@quartzbio.com."
        )
        err = "A %s was raised" % (type(e).__name__,)

        if isinstance(e, (urllib3.exceptions.SSLError, ssl.SSLError)) or sys.version_info >= (3, 12):
            err += ("\n\nThis is an SSLError. If you're using python 3.12, "
                    "it could be because of stricter SSL requirements:\n"
                    "https://docs.python.org/3/whatsnew/3.12.html\n"
                    "https://docs.python.org/3/whatsnew/3.12.html\n"
                    "Try upgrading urllib3 and certifi:\n"
                    "  pip install --upgrade urllib3 certifi\n\n")
        elif str(e):
            err += " with error message %s" % (str(e),)
        else:
            err += " with no error message"
    msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
    raise QuartzBioError(message=msg)


class QuartzBioClient(object):
    """A requests-based HTTP client for QuartzBio API resources"""

    def __init__(
        self,
        host=None,
        token=None,
        token_type: Literal["Bearer", "Token"] = "Token",
        include_resources=True,
        retry_all: bool = None,
    ):
        self._host: str = None
        self._auth: QuartzBioTokenAuth = None
        self._session: Session = None

        self._headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Accept-Encoding": "gzip,deflate",
        }
        self.retry_all = bool(retry_all)
        if self.retry_all is None:
            self.retry_all = bool(os.environ.get("QUARTZBIO_RETRY_ALL"))

        # this class is created before any commands, so it shouldn't raise a missing host exception
        self.set_credentials(host, token, token_type, raise_on_missing=False)
        self.set_user_agent()

        # Import all resources into the client
        if include_resources:
            skip = (
                "QuartzBioError",
                "QuartzBioClient",
            )
            for name, class_ in inspect.getmembers(quartzbio, inspect.isclass):
                if name in skip:
                    continue
                subclass = type(name, (class_,), {"_client": self})
                setattr(self, name, subclass)

    def set_user_agent(self, name=None, version=None):
        ua = "quartzbio-python-client/{} python-requests/{} {}/{}".format(
            VERSION,
            requests.__version__,
            platform.python_implementation(),
            platform.python_version(),
        )

        # Prefix the name of the app or script before the
        # default user-agent.
        if name:
            name = name.replace(" ", "-")
            if version:
                ua = "{}/{} {}".format(name, version, ua)
            else:
                ua = "{} {}".format(name, ua)

        self._headers["User-Agent"] = ua

    def set_credentials(
        self, host: str, token: str, token_type: Literal["Bearer", "Token"],
        *, debug: bool = False, raise_on_missing: bool = True
    ):
        self._host, self._auth = authenticate(
            host, token, token_type, debug=debug, raise_on_missing=raise_on_missing
        )

        if self._host:
            retry_kwargs = {}
            if self.retry_all:
                logger.info("Retries enabled for all API requests")
                retry_kwargs["allow_redirects"] = frozenset(
                    ["HEAD", "GET", "PUT", "POST", "PATCH", "DELETE", "OPTIONS", "TRACE"]
                )
            else:
                logger.info("Retries enabled for read-only API requests")

            retries = Retry(
                total=5,
                backoff_factor=2,
                status_forcelist=[
                    codes.bad_gateway,
                    codes.service_unavailable,
                    codes.gateway_timeout,
                ],
                **retry_kwargs
            )

            # Use a session with a retry policy to handle
            # intermittent connection errors.
            adapter = adapters.HTTPAdapter(max_retries=retries)
            self._session = Session()
            self._session.mount(self._host, adapter)

    def is_logged_in(self):
        return bool(self._host and self._auth and self._session)

    def whoami(self):
        return self.get("/v1/user", {})

    def get(self, url, params, **kwargs):
        """Issues an HTTP GET across the wire via the Python requests
        library. See *request()* for information on keyword args."""
        kwargs["params"] = params
        return self.request("GET", url, **kwargs)

    def post(self, url, data, **kwargs):
        """Issues an HTTP POST across the wire via the Python requests
        library. See *request* for information on keyword args."""
        kwargs["data"] = data
        return self.request("POST", url, **kwargs)

    def delete(self, url, data, **kwargs):
        """Issues an HTTP DELETE across the wire via the Python requests
        library. See *request* for information on keyword args."""
        kwargs["data"] = data
        return self.request("DELETE", url, **kwargs)

    def request(self, method, url, **kwargs):
        """
        Issues an HTTP Request across the wire via the Python requests
        library.

        Parameters
        ----------

        method : str
        an HTTP method: GET, PUT, POST, DELETE, ...

        url : str
        the place to connect to. If the url doesn't start
        with a protocol (https:// or http://), we'll slap
        quartzbio.api_host in the front.

        allow_redirects: bool, optional
        set *False* we won't follow any redirects

        headers: dict, optional

          Custom headers can be provided here; generally though this
          will be set correctly by default dependent on the
          method type. If the content type is JSON, we'll
          JSON-encode params.

        param : dict, optional
           passed as *params* in the requests.request

        timeout : int, optional
          timeout value in seconds for the request

        raw: bool, optional
          unless *True* the response encoded to json

        files: file
          File content in the form of a file handle which is to be
          uploaded. Files are passed in POST requests

        Returns
        -------
        response object. If *raw* is not *True* and
        repsonse if valid the object will be JSON encoded. Otherwise
        it will be the request.reposne object.
        """
        if not self.is_logged_in():
            raise QuartzBioError("HTTP request: client is not logged in!")

        opts = {
            "allow_redirects": True,
            "auth": self._auth,
            "data": {},
            "files": None,
            "headers": dict(self._headers),
            "params": {},
            "timeout": 80,
            "verify": True,
        }

        raw = kwargs.pop("raw", False)
        debug = kwargs.pop("debug", False)
        opts.update(kwargs)
        method = method.upper()

        if opts["files"]:
            # Don't use application/json for file uploads or GET requests
            opts["headers"].pop("Content-Type", None)
        else:
            opts["data"] = json.dumps(opts["data"])

        if not url.startswith(self._host):
            url = urljoin(self._host, url)

        logger.debug("API %s Request: %s" % (method, url))

        if debug:
            self._log_raw_request(method, url, **opts)

        try:
            response = self._session.request(method, url, **opts)
        except Exception as e:
            _handle_request_error(e)

        if 429 == response.status_code:
            delay = int(response.headers["retry-after"]) + 1
            logger.warn("Too many requests. Retrying in {0}s.".format(delay))
            time.sleep(delay)
            return self.request(method, url, **kwargs)

        if not (200 <= response.status_code < 400):
            _handle_api_error(response)

        # 204 is used on deletion. There is no JSON here.
        if raw or response.status_code in [204, 301, 302]:
            return response

        try:
            return response.json()
        except Exception:
            if b"<!DOCTYPE html>" in response.content[:40]:
                helper_msg = self._host
                if suggested_host := self.validate_host_is_www_url(self._host):
                    helper_msg = f"Provided API host is: `{self._host}`. Did you perhaps mean `{suggested_host}`?"

                html_response = response.text[:30].replace("\n", "")
                raise QuartzBioError(
                    f"QuartzBio error: HTML response received from API. {html_response}...\n"
                    "  Please confirm that API Host doesn't point to EDP's Web URL:\n"
                    f"  {helper_msg}"
                )
            elif self._host is None:
                # shouldn't happen, maybe in rare race conditions
                raise QuartzBioError("No EDP API host was set!")

            raise QuartzBioError(
                f"Could not parse JSON response: {response.content}"
            )

    def validate_host_is_www_url(self, host):
        # returns corrected API's URL, if it suspects that host is a WWW url, not API
        if ".api.edp" not in host and ".edp" in host:
            return host.replace(".edp", ".api.edp")

    def _log_raw_request(self, method, url, **kwargs):
        from requests import Request, Session

        req = Request(
            method=method.upper(), url=url, data=kwargs["data"], params=kwargs["params"]
        )
        prepped = Session().prepare_request(
            req,
        )
        logger.debug(prepped.headers)
        logger.debug(prepped.body)

    def __repr__(self):
        return "<QuartzBioClient {0} {1}>".format(self._host, self._auth)


client = QuartzBioClient(include_resources=False)



import os
import time

from ..client import client
from ..query import Query

from .quartzbio_object import convert_to_quartzbio_object
from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource

from .task import Task
from .datasetfield import DatasetField
from .datasetexport import DatasetExport
from .datasetmigration import DatasetMigration


class Dataset(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    UpdateableAPIResource,
):
    """
    Datasets are access points to data. Dataset names are unique
    within a vault folder.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("vault_name", "Vault"),
        ("vault_object_path", "Path"),
        ("documents_count", "Documents"),
        ("description", "Description"),
    )

    @classmethod
    def make_full_path(cls, vault_name, path, name, **kwargs):
        from quartzbio import QuartzBioError

        _client = kwargs.pop("client", None) or cls._client or client

        try:
            user = _client.get("/v1/user", {})
            domain = user["account"]["domain"]
        except QuartzBioError as e:
            print("Error obtaining account domain: {0}".format(e))
            raise

        result = ":".join(
            [
                domain,
                vault_name,
                os.path.join("/" + path.lstrip("/"), name),
            ]
        )

        return result

    @classmethod
    def get_by_full_path(cls, full_path, **kwargs):
        from quartzbio import Object

        _client = kwargs.pop("client", None) or cls._client or client
        obj = Object.get_by_full_path(full_path, assert_type="dataset", client=_client)
        return Dataset.retrieve(obj["dataset_id"], client=_client, **kwargs)

    @classmethod
    def get_or_create_by_full_path(cls, full_path, **kwargs):
        from quartzbio import Object

        # Assert this is a dataset (for the "get" in get_or_create)
        kwargs["assert_type"] = "dataset"
        # Create this is a dataset (for the "create" in get_or_create)
        kwargs["object_type"] = "dataset"

        _client = kwargs.pop("client", None) or cls._client or client
        obj = Object.get_or_create_by_full_path(full_path, client=_client, **kwargs)

        return cls.retrieve(obj.dataset_id)

    def saved_queries(self, **params):
        from quartzbio import SavedQuery

        if "id" not in self or not self["id"]:
            raise Exception(
                "No Dataset ID was provided. "
                "Please instantiate the Dataset "
                "object with an ID."
            )

        saved_queries = SavedQuery.all(
            dataset_id=self["id"], client=self._client, **params
        )

        return saved_queries

    def fields(self, name=None, **params):
        if "fields_url" not in self:
            # Dataset object may not have been retrieved. Grab it.
            self.refresh()

        if name:
            params.update(
                {
                    "name": name,
                }
            )
            fields = self._client.get(self.fields_url, params)
            result = DatasetField.retrieve(fields["data"][0]["id"], client=self._client)
            return result

        response = self._client.get(self.fields_url, params)
        results = convert_to_quartzbio_object(response, client=self._client)
        results.set_tabulate(
            ["name", "data_type", "entity_type", "description"],
            headers=["Field", "Data Type", "Entity Type", "Description"],
            aligns=["left", "left", "left", "left"],
            sort=True,
        )

        return results

    def template(self, **params):
        if "template_url" not in self:
            # Dataset object may not have been retrieved. Grab it.
            self.refresh()

        response = self._client.get(self.template_url, params)
        return convert_to_quartzbio_object(response, client=self._client)

    def commits(self, **params):
        if "commits_url" not in self:
            # Dataset object may not have been retrieved. Grab it.
            self.refresh()

        response = self._client.get(self.commits_url, params)
        results = convert_to_quartzbio_object(response, client=self._client)
        results.set_tabulate(
            ["id", "title", "description", "status", "created_at"],
            headers=["ID", "Title", "Description", "Status", "Created"],
            aligns=["left", "left", "left", "left", "left"],
            sort=False,
        )

        return results

    def imports(self, **params):
        if "imports_url" not in self:
            # Dataset object may not have been retrieved. Grab it.
            self.refresh()

        response = self._client.get(self.imports_url, params)
        results = convert_to_quartzbio_object(response, client=self._client)
        results.set_tabulate(
            ["id", "title", "description", "status", "created_at"],
            headers=["ID", "Title", "Description", "Status", "Created"],
            aligns=["left", "left", "left", "left", "left"],
            sort=False,
        )

        return results

    def _data_url(self):
        if "data_url" not in self:
            if "id" not in self or not self["id"]:
                raise Exception(
                    "No Dataset ID was provided. "
                    "Please instantiate the Dataset "
                    "object with an ID."
                )
            # automatically construct the data_url from the ID
            return self.instance_url() + "/data"
        return self["data_url"]

    def query(self, query=None, **params):
        self._data_url()  # raises an exception if there's no ID
        return Query(self["id"], query=query, client=self._client, **params)

    def lookup(self, *sbids):
        lookup_url = self._data_url() + "/" + ",".join(sbids)
        return self._client.get(lookup_url, {})["results"]

    def _beacon_url(self):
        if "beacon_url" not in self:
            if "id" not in self or not self["id"]:
                raise Exception(
                    "No Dataset ID was provided. "
                    "Please instantiate the Dataset "
                    "object with an ID."
                )
            # automatically construct the data_url from the ID
            self["beacon_url"] = self.instance_url() + "/beacon"
        return self["beacon_url"]

    def beacon(self, **params):
        # raises an exception if there's no ID
        return self._client.get(self._beacon_url(), params)

    def import_file(self, path, **kwargs):
        """
        This is a shortcut to creating a DatasetImport. Can't use "import()"
        because of Python.
        """
        from . import Manifest
        from . import DatasetImport

        if "id" not in self or not self["id"]:
            raise Exception(
                "No Dataset ID found. "
                "Please instantiate or retrieve a dataset "
                "with an ID."
            )

        manifest = Manifest()
        manifest.add(path)
        return DatasetImport.create(
            dataset_id=self["id"], manifest=manifest.manifest, **kwargs
        )

    def export(self, format="json", follow=True, **kwargs):
        if "id" not in self or not self["id"]:
            raise Exception(
                "No Dataset ID was provided. "
                "Please instantiate the Dataset "
                "object with an ID."
            )

        export = DatasetExport.create(dataset_id=self["id"], format=format, **kwargs)

        if follow:
            export.follow()

        return export

    def migrate(self, target, follow=True, **kwargs):
        """
        Migrate the data from this dataset to a target dataset.

        Valid optional kwargs include:

        * source_params
        * target_fields
        * include_errors
        * commit_mode

        """
        if "id" not in self or not self["id"]:
            raise Exception(
                "No source dataset ID found. "
                "Please instantiate the Dataset "
                "object with an ID."
            )

        # Target can be provided as an object or as an ID.
        try:
            target_id = target.id
        except AttributeError:
            target_id = target

        migration_resp = DatasetMigration.create(
            source_id=self["id"], target_id=target_id, **kwargs
        )

        if follow:
            # If migration was created with the parallel flag, then multiple
            # migration objects will be returned.
            if "data" in migration_resp:
                for migration in migration_resp["data"]:
                    migration.follow()
            else:
                migration_resp.follow()

        return migration_resp

    def archive(self, storage_class=None, follow=False):
        """
        Archive this dataset
        """
        return self.vault_object.archive(storage_class=storage_class, follow=follow)

    def restore(self, storage_class=None, follow=False, **kwargs):
        """
        Restore this dataset
        """
        return self.vault_object.restore(storage_class=storage_class, follow=follow)

    def activity(self, follow=False, limit=1, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        """Get a list of active Tasks that have a target object of
        the dataset. Active tasks are in the running, queued or pending state.

        Defaults to limit=1 for performance. Increase this value in order
        to return more tasks as output.
        """
        statuses = ["running", "queued", "pending"]

        while True:
            # NOTE: source_object_id is not being queried here
            # and therefore active DatasetExports will not appear
            # in activity
            activity = Task.all(
                target_object_id=self.id,
                status=",".join(statuses),
                limit=limit,
                ordering="-updated_at",
                client=self._client,
            )

            print("Found {0} active task(s)".format(activity.total))
            if not activity or not follow:
                break

            for task in activity.quartzbio_objects()[0:limit]:
                task.follow(sleep_seconds=sleep_seconds)

            time.sleep(sleep_seconds)

        return list(activity)

    #
    # Vault properties
    #
    @property
    def vault_object(self):
        from quartzbio import Object

        return Object.retrieve(self["vault_object_id"], client=self._client)

    def enable_global_beacon(self):
        """
        Enable Global Beacon for this dataset.
        """
        return self.vault_object.enable_global_beacon()

    def disable_global_beacon(self):
        """
        Disable Global Beacon for this dataset.
        """
        return self.vault_object.disable_global_beacon()

    def get_global_beacon_status(self, raise_on_disabled=False):
        """
        Retrieves the Global Beacon status for this dataset.
        """
        return self.vault_object.get_global_beacon_status(raise_on_disabled)



"""QuartzBio Vault API resource"""

from ..client import client
from ..errors import NotFoundError

from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import SearchableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource

import re


class Vault(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    SearchableAPIResource,
    UpdateableAPIResource,
):
    """
    A vault is like a filesystem that can contain files, folder,
    and QuartzBio datasets.  Vaults can be "connected" to external resources
    such as Amazon S3 Buckets.
    Typically, vaults contain a series of datasets that are compatible with
    each other (i.e. they come from the same data source or project).
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("full_path", "Full Path"),
        ("provider", "Provider"),
        ("description", "Description"),
    )

    # Regex describing Vault full path.
    # NOTE: Not valid for object full paths.
    VAULT_PATH_RE = re.compile(
        # Non-greedy wildcard match for domain
        r"^(?:(?P<domain>[a-zA-Z0-9\-]+)\:)??"
        # Match vault or vault:/ or vault/
        r"(?P<vault>[^\/:]+)(?:\:?\/.*)?$"
    )

    def __init__(self, vault_id, **kwargs):
        super(Vault, self).__init__(vault_id, **kwargs)

    def _object_list_helper(self, **params):
        from quartzbio import Object

        params.update(
            {
                "vault_id": self.id,
            }
        )

        items = Object.all(client=self._client, **params)
        return items

    @classmethod
    def validate_full_path(cls, full_path, **kwargs):
        """Helper method to return a full path from a full or partial path.

            If no domain, assumes user's account domain
            If the vault is "~", assumes personal vault.

        Valid vault paths include:

            domain:vault
            domain:vault:/path
            domain:vault/path
            vault:/path
            vault
            ~/

        Invalid vault paths include:

            /vault/
            /path
            /
            :/

        Does not allow overrides for any vault path components.
        """
        _client = kwargs.pop("client", None) or cls._client or client

        full_path = full_path.strip()
        if not full_path:
            raise Exception(
                'Vault path "{0}" is invalid. Path must be in the format: '
                '"domain:vault:/path" or "vault:/path".'.format(full_path)
            )

        match = cls.VAULT_PATH_RE.match(full_path)
        if not match:
            raise Exception(
                'Vault path "{0}" is invalid. Path must be in the format: '
                '"domain:vault:/path" or "vault:/path".'.format(full_path)
            )
        path_parts = match.groupdict()

        # Handle the special case where "~" means personal vault
        if path_parts.get("vault") == "~":
            path_parts = dict(domain=None, vault=None)

        # If any values are None, set defaults from the user.
        if None in path_parts.values():
            user = _client.get("/v1/user", {})
            defaults = {
                "domain": user["account"]["domain"],
                "vault": "user-{0}".format(user["id"]),
            }
            path_parts = dict((k, v or defaults.get(k)) for k, v in path_parts.items())

        # Rebuild the full path
        full_path = "{domain}:{vault}".format(**path_parts)
        path_parts["vault_full_path"] = full_path
        return full_path, path_parts

    def files(self, **params):
        return self._object_list_helper(object_type="file", **params)

    def folders(self, **params):
        return self._object_list_helper(object_type="folder", **params)

    def datasets(self, **params):
        return self._object_list_helper(object_type="dataset", **params)

    def objects(self, **params):
        return self._object_list_helper(**params)

    def ls(self, **params):
        return self._object_list_helper(**params)

    def _get_parent_folder(self, path):
        from quartzbio import Object

        return Object.get_by_full_path(
            ":".join([self.full_path, path]), assert_type="folder", client=self._client
        )

    def create_dataset(self, name, **params):
        from quartzbio import Object

        params["vault_id"] = self.id
        params["object_type"] = "dataset"
        path = params.pop("path", None)

        if path == "/" or path is None:
            params["parent_object_id"] = None
        else:
            parent_object = self._get_parent_folder(path)
            params["parent_object_id"] = parent_object.id

        params["filename"] = name
        return Object.create(**params)

    def create_folder(self, filename, **params):
        from quartzbio import Object

        path = params.pop("path", None)
        if path and path != "/":
            parent_object = self._get_parent_folder(path)
            params["parent_object_id"] = parent_object.id

        params.update(
            {"filename": filename, "vault_id": self.id, "object_type": "folder"}
        )
        return Object.create(client=self._client, **params)

    def upload_file(self, local_path, remote_path, **kwargs):
        from quartzbio import Object

        return Object.upload_file(local_path, remote_path, self.full_path, client=self._client, **kwargs)

    def search(self, query, **params):
        params.update(
            {
                "query": query,
            }
        )
        return self._object_list_helper(**params)

    def enable_versioning(self):
        self["versioning"] = "enabled"
        self.save()

    def disable_versioning(self):
        self["versioning"] = "disabled"
        self.save()

    def suspend_versioning(self):
        self["versioning"] = "suspended"
        self.save()

    @classmethod
    def get_by_full_path(cls, full_path, **kwargs):
        _client = kwargs.pop("client", None) or cls._client or client

        full_path, parts = cls.validate_full_path(full_path, client=_client)
        return Vault._retrieve_helper(
            "vault",
            "name",
            full_path,
            account_domain=parts["domain"],
            name=parts["vault"],
            client=_client,
        )

    @classmethod
    def get_or_create_by_full_path(cls, full_path, **kwargs):
        _client = kwargs.pop("client", None) or cls._client or client

        try:
            return Vault.get_by_full_path(full_path, client=_client)
        except NotFoundError:
            pass

        # Vault not found, create it
        parts = full_path.split(":", 2)
        vault_name = parts[-1]

        return Vault.create(name=vault_name, client=_client)

    @classmethod
    def get_personal_vault(cls, **kwargs):
        _client = kwargs.pop("client", None) or cls._client or client
        user = _client.get("/v1/user", {})
        # TODO - this will have to change if the format of the personal vaults
        # changes.
        name = "user-{0}".format(user["id"])
        return list(Vault.all(name=name, vault_type="user", client=_client))[0]

    @classmethod
    def get_or_create_uploads_path(cls, **kwargs):
        from quartzbio import Object

        _client = kwargs.pop("client", None) or cls._client or client
        v = cls.get_personal_vault(client=_client)
        default_path = "Uploads"
        full_path = "{0}:/{1}".format(v.full_path, default_path)

        try:
            upload_dir = Object.get_by_full_path(
                full_path, assert_type="folder", client=_client
            )
        except NotFoundError:
            print("Uploads directory not found. Creating {0}".format(full_path))
            upload_dir = Object.create(
                vault_id=v.id,
                object_type="folder",
                filename=default_path,
                client=_client,
            )

        return upload_dir.path



# -*- coding: utf-8 -*-
#
# This file contains code from python-tabulate, modified for QuartzBio
#
# Copyright © 2011-2013 Sergey Astanin
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from collections import namedtuple
import re
from .printing import TTY_COLS
from itertools import zip_longest
from functools import reduce

_none_type = type(None)
_int_type = int
_float_type = float
_text_type = str
_binary_type = bytes


__all__ = ["tabulate"]
__version__ = "0.6"


Line = namedtuple("Line", ["begin", "hline", "sep", "end"])
DataRow = namedtuple("DataRow", ["begin", "sep", "end"])
TableFormat = namedtuple(
    "TableFormat",
    [
        "lineabove",
        "linebelowheader",
        "linebetweenrows",
        "linebelow",
        "headerrow",
        "datarow",
        "padding",
        "usecolons",
        "with_header_hide",
        "without_header_hide",
    ],
)
_format_defaults = {
    "padding": 0,
    "usecolons": False,
    "with_header_hide": [],
    "without_header_hide": [],
}

_table_formats = {
    "simple": TableFormat(
        lineabove=None,
        linebelowheader=Line("", "-", "  ", ""),
        linebetweenrows=None,
        linebelow=Line("", "-", "  ", ""),
        headerrow=DataRow("", "  ", ""),
        datarow=DataRow("", "  ", ""),
        padding=0,
        usecolons=False,
        with_header_hide=["linebelow"],
        without_header_hide=[],
    ),
    "plain": TableFormat(
        None,
        None,
        None,
        None,
        DataRow("", "  ", ""),
        DataRow("", "  ", ""),
        **_format_defaults,
    ),
    "grid": TableFormat(
        lineabove=Line("+", "-", "+", "+"),
        linebelowheader=Line("+", "=", "+", "+"),
        linebetweenrows=Line("+", "-", "+", "+"),
        linebelow=Line("+", "-", "+", "+"),
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        usecolons=False,
        with_header_hide=[],
        without_header_hide=["linebelowheader"],
    ),
    "pipe": TableFormat(
        lineabove=None,
        linebelowheader=Line("|", "-", "|", "|"),
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        usecolons=True,
        with_header_hide=[],
        without_header_hide=[],
    ),
    "orgmode": TableFormat(
        lineabove=None,
        linebelowheader=Line("|", "-", "+", "|"),
        linebetweenrows=None,
        linebelow=None,
        headerrow=DataRow("|", "|", "|"),
        datarow=DataRow("|", "|", "|"),
        padding=1,
        usecolons=False,
        with_header_hide=[],
        without_header_hide=["linebelowheader"],
    ),
}


_invisible_codes = re.compile(r"\x1b\[\d*m")  # ANSI color codes


def simple_separated_format(separator):
    """
    Construct a simple TableFormat with columns separated by a separator.

    >>> tsv = simple_separated_format("\t") ; \
        tabulate([["foo", 1], ["spam", 23]], \
            tablefmt=tsv) == 'foo \\t 1\\nspam\\t23'
    True

    """
    return TableFormat(
        None,
        None,
        None,
        None,
        headerrow=None,
        datarow=DataRow("", separator, ""),
        **_format_defaults,
    )


def _isconvertible(conv, string):
    try:
        conv(string)  # noqa
        return True
    except (TypeError, ValueError):
        return False


def _isnumber(string):
    """
    >>> _isnumber("123.45")
    True
    >>> _isnumber("123")
    True
    >>> _isnumber("spam")
    False
    """
    return _isconvertible(float, string)


def _isint(string):
    """
    >>> _isint("123")
    True
    >>> _isint("123.45")
    False
    """
    return (
        type(string) is int
        or (isinstance(string, _binary_type) or isinstance(string, str))
        and _isconvertible(int, string)
    )


def _type(string, has_invisible=True):
    """
    The least generic type (type(None), int, float, str, unicode).

    >>> _type(None) is type(None)
    True
    >>> _type("foo") is type("")
    True
    >>> _type("1") is type(1)
    True
    >>> _type('\x1b[31m42\x1b[0m') is type(42)
    True
    >>> _type('\x1b[31m42\x1b[0m') is type(42)
    True

    """

    if has_invisible and (
        isinstance(string, _text_type) or isinstance(string, _binary_type)
    ):
        string = _strip_invisible(string)

    if string is None:
        return _none_type
    elif _isint(string):
        return _int_type
    elif _isnumber(string):
        return float
    elif isinstance(string, _binary_type):
        return _binary_type
    else:
        return _text_type


def _afterpoint(string):
    """
    Symbols after a decimal point, -1 if the string lacks the decimal point.

    >>> _afterpoint("123.45")
    2
    >>> _afterpoint("1001")
    -1
    >>> _afterpoint("eggs")
    -1
    >>> _afterpoint("123e45")
    2

    """
    if _isnumber(string):
        if _isint(string):
            return -1
        else:
            pos = string.rfind(".")
            pos = string.lower().rfind("e") if pos < 0 else pos
            if pos >= 0:
                return len(string) - pos - 1
            else:
                return -1  # no point
    else:
        return -1  # not a number


def _padleft(width, s, has_invisible=True):
    """
    Flush right.

    >>> _padleft(6, '\u044f\u0439\u0446\u0430') \
        == '  \u044f\u0439\u0446\u0430'
    True

    """
    iwidth = width + len(s) - len(_strip_invisible(s)) if has_invisible else width
    fmt = "{0:>%ds}" % iwidth
    return fmt.format(s)


def _padright(width, s, has_invisible=True):
    """
    Flush left.

    >>> _padright(6, '\u044f\u0439\u0446\u0430') \
        == '\u044f\u0439\u0446\u0430  '
    True

    """
    iwidth = width + len(s) - len(_strip_invisible(s)) if has_invisible else width
    fmt = "{0:<%ds}" % iwidth
    return fmt.format(s)


def _padboth(width, s, has_invisible=True):
    """
    Center string.

    >>> _padboth(6, '\u044f\u0439\u0446\u0430') \
        == ' \u044f\u0439\u0446\u0430 '
    True

    """
    iwidth = width + len(s) - len(_strip_invisible(s)) if has_invisible else width
    fmt = "{0:^%ds}" % iwidth
    return fmt.format(s)


def _strip_invisible(s):
    "Remove invisible ANSI color codes."
    return re.sub(_invisible_codes, "", s)


def _visible_width(s):
    """
    Visible width of a printed string. ANSI color codes are removed.

    >>> _visible_width('\x1b[31mhello\x1b[0m'), _visible_width("world")
    (5, 5)

    """
    if isinstance(s, _text_type) or isinstance(s, _binary_type):
        return len(_strip_invisible(s))
    else:
        return len(_text_type(s))


def _align_column(strings, alignment, minwidth=0, has_invisible=True):
    """
    [string] -> [padded_string]

    >>> list(map(str,_align_column( \
        ["12.345", "-1234.5", "1.23", "1234.5", \
         "1e+234", "1.0e234"], "decimal")))
    ['   12.345  ', '-1234.5    ', '    1.23   ', \
     ' 1234.5    ', '    1e+234 ', '    1.0e234']

    """
    if alignment == "right":
        strings = [s.strip() for s in strings]
        padfn = _padleft
    elif alignment in "center":
        strings = [s.strip() for s in strings]
        padfn = _padboth
    elif alignment in "decimal":
        decimals = [_afterpoint(s) for s in strings]
        maxdecimals = max(decimals)
        strings = [s + (maxdecimals - decs) * " " for s, decs in zip(strings, decimals)]
        padfn = _padleft
    else:
        strings = [s.strip() for s in strings]
        padfn = _padright

    if has_invisible:
        width_fn = _visible_width
    else:
        width_fn = len

    maxwidth = max(max(list(map(width_fn, strings))), minwidth)
    padded_strings = [padfn(maxwidth, s, has_invisible) for s in strings]
    return padded_strings


def _more_generic(type1, type2):
    types = {_none_type: 0, int: 1, float: 2, _text_type: 4}
    invtypes = {4: _text_type, 2: float, 1: int, 0: _none_type}
    moregeneric = max(types.get(type1, 4), types.get(type2, 4))
    return invtypes[moregeneric]


def _column_type(strings, has_invisible=True):
    """
    The least generic type all column values are convertible to.

    >>> _column_type(["1", "2"]) is _int_type
    True
    >>> _column_type(["1", "2.3"]) is _float_type
    True
    >>> _column_type(["1", "2.3", "four"]) is _text_type
    True
    >>> _column_type(["four", '\u043f\u044f\u0442\u044c']) is _text_type
    True
    >>> _column_type([None, "brux"]) is _text_type
    True
    >>> _column_type([1, 2, None]) is _int_type
    True

    """
    types = [_type(s, has_invisible) for s in strings]
    return reduce(_more_generic, types, int)


def _format(val, valtype, floatfmt, missingval=""):
    """
    Format a value accoding to its type.

    Unicode is supported:

    >>> hrow = ['\u0431\u0443\u043a\u0432\u0430', \
                '\u0446\u0438\u0444\u0440\u0430'] ; \
        tbl = [['\u0430\u0437', 2], ['\u0431\u0443\u043a\u0438', 4]] ; \
        good_result = '\\u0431\\u0443\\u043a\\u0432\\u0430      \
                        \\u0446\\u0438\\u0444\\u0440\\u0430\\n-------\
                          -------\\n\\u0430\\u0437             \
                          2\\n\\u0431\\u0443\\u043a\\u0438           4' ; \
        tabulate(tbl, headers=hrow) == good_result
    True

    """
    if val is None:
        return missingval

    if valtype in [int, _binary_type, _text_type]:
        return "{0}".format(val)
    elif valtype is float:
        return format(float(val), floatfmt)
    else:
        return "{0}".format(val)


def _align_header(header, alignment, width):
    if alignment == "left":
        return _padright(width, header)
    elif alignment == "center":
        return _padboth(width, header)
    else:
        return _padleft(width, header)


def _normalize_tabular_data(tabular_data, headers, sort=True):
    """
    Transform a supported data type to a list of lists, and a list of headers.

    Supported tabular data types:

    * list-of-lists or another iterable of iterables

    * 2D NumPy arrays

    * dict of iterables (usually used with headers="keys")

    * pandas.DataFrame (usually used with headers="keys")

    The first row can be used as headers if headers="firstrow",
    column indices can be used as headers if headers="keys".

    """

    if hasattr(tabular_data, "keys") and hasattr(tabular_data, "values"):
        # dict-like and pandas.DataFrame?
        if hasattr(tabular_data.values, "__call__"):
            # likely a conventional dict
            keys = list(tabular_data.keys())
            # columns have to be transposed
            rows = list(zip_longest(*list(tabular_data.values())))
        elif hasattr(tabular_data, "index"):
            # values is a property, has .index then
            # it's likely a pandas.DataFrame (pandas 0.11.0)
            keys = list(tabular_data.keys())
            # values matrix doesn't need to be transposed
            vals = tabular_data.values
            names = tabular_data.index
            rows = [[v] + list(row) for v, row in zip(names, vals)]
        else:
            raise ValueError(
                "tabular data doesn't appear to be a dict " "or a DataFrame"
            )

        if headers == "keys":
            headers = list(map(_text_type, keys))  # headers should be strings

    else:  # it's, as usual, an iterable of iterables, or a NumPy array
        rows = list(tabular_data)

        if headers == "keys" and len(rows) > 0:  # keys are column indices
            headers = list(map(_text_type, list(range(len(rows[0])))))

    # take headers from the first row if necessary
    if headers == "firstrow" and len(rows) > 0:
        headers = list(map(_text_type, rows[0]))  # headers should be strings
        rows = rows[1:]

    headers = list(headers)

    rows = list(map(list, rows))

    if sort and len(rows) > 1:
        rows = sorted(rows, key=lambda x: x[0])

    # pad with empty headers for initial columns if necessary
    if headers and len(rows) > 0:
        nhs = len(headers)
        ncols = len(rows[0])
        if nhs < ncols:
            headers = [""] * (ncols - nhs) + headers

    return rows, headers


def _build_row(cells, padding, begin, sep, end):
    "Return a string which represents a row of data cells."

    pad = " " * padding
    padded_cells = [pad + cell + pad for cell in cells]

    # QuartzBio: we're only displaying Key-Value tuples (dimension of 2).
    #  enforce that we don't wrap lines by setting a max
    #  limit on row width which is equal to TTY_COLS (see printing)
    rendered_cells = (begin + sep.join(padded_cells) + end).rstrip()
    if len(rendered_cells) > TTY_COLS:
        if not cells[-1].endswith(" ") and not cells[-1].endswith("-"):
            terminating_str = " ... "
        else:
            terminating_str = ""
        rendered_cells = "{0}{1}{2}".format(
            rendered_cells[: TTY_COLS - len(terminating_str) - 1], terminating_str, end
        )

    return rendered_cells


def _build_line(colwidths, padding, begin, fill, sep, end):
    "Return a string which represents a horizontal line."
    cells = [fill * (w + 2 * padding) for w in colwidths]
    return _build_row(cells, 0, begin, sep, end)


def _mediawiki_cell_attrs(row, colaligns):
    "Prefix every cell in a row with an HTML alignment attribute."
    alignment = {
        "left": "",
        "right": 'align="right"| ',
        "center": 'align="center"| ',
        "decimal": 'align="right"| ',
    }
    row2 = [alignment[a] + c for c, a in zip(row, colaligns)]
    return row2


def _line_segment_with_colons(linefmt, align, colwidth):
    """Return a segment of a horizontal line with optional colons which
    indicate column's alignment (as in `pipe` output format)."""
    fill = linefmt.hline
    w = colwidth
    if align in ["right", "decimal"]:
        return (fill[0] * (w - 1)) + ":"
    elif align == "center":
        return ":" + (fill[0] * (w - 2)) + ":"
    elif align == "left":
        return ":" + (fill[0] * (w - 1))
    else:
        return fill[0] * w


def _format_table(fmt, headers, rows, colwidths, colaligns):
    """Produce a plain-text representation of the table."""
    lines = []
    hidden = fmt.with_header_hide if headers else fmt.without_header_hide
    pad = fmt.padding
    headerrow = fmt.headerrow if fmt.headerrow else fmt.datarow

    if fmt.lineabove and "lineabove" not in hidden:
        lines.append(_build_line(colwidths, pad, *fmt.lineabove))

    if headers:
        lines.append(_build_row(headers, pad, *headerrow))

    if fmt.linebelowheader and "linebelowheader" not in hidden:
        begin, fill, sep, end = fmt.linebelowheader
        if fmt.usecolons:
            segs = [
                _line_segment_with_colons(fmt.linebelowheader, a, w + 2 * pad)
                for w, a in zip(colwidths, colaligns)
            ]
            lines.append(_build_row(segs, 0, begin, sep, end))
        else:
            lines.append(_build_line(colwidths, pad, *fmt.linebelowheader))

    if rows and fmt.linebetweenrows and "linebetweenrows" not in hidden:
        # initial rows with a line below
        for row in rows[:-1]:
            lines.append(_build_row(row, pad, *fmt.datarow))
            lines.append(_build_line(colwidths, pad, *fmt.linebetweenrows))
        # the last row without a line below
        lines.append(_build_row(rows[-1], pad, *fmt.datarow))
    else:
        for row in rows:
            lines.append(_build_row(row, pad, *fmt.datarow))

    if fmt.linebelow and "linebelow" not in hidden:
        lines.append(_build_line(colwidths, pad, *fmt.linebelow))

    return "\n".join(lines)


def tabulate(
    tabular_data,
    headers=[],
    tablefmt="orgmode",
    floatfmt="g",
    aligns=[],
    missingval="",
    sort=True,
    is_tsv=False,
):
    list_of_lists, headers = _normalize_tabular_data(tabular_data, headers, sort=sort)

    # optimization: look for ANSI control codes once,
    # enable smart width functions only if a control code is found
    plain_text = "\n".join(
        ["\t".join(map(_text_type, headers))]
        + ["\t".join(map(_text_type, row)) for row in list_of_lists]
    )

    has_invisible = re.search(_invisible_codes, plain_text)
    if has_invisible:
        width_fn = _visible_width
    else:
        width_fn = len

    # format rows and columns, convert numeric values to strings
    cols = list(zip(*list_of_lists))

    coltypes = list(map(_column_type, cols))
    cols = [
        [_format(v, ct, floatfmt, missingval) for v in c]
        for c, ct in zip(cols, coltypes)
    ]

    # align columns
    if not aligns:
        # dynamic alignment by col type
        aligns = ["decimal" if ct in [int, float] else "left" for ct in coltypes]

    minwidths = [width_fn(h) + 2 for h in headers] if headers else [0] * len(cols)
    cols = [
        _align_column(c, a, minw, has_invisible)
        for c, a, minw in zip(cols, aligns, minwidths)
    ]

    if headers:
        # align headers and add headers
        minwidths = [max(minw, width_fn(c[0])) for minw, c in zip(minwidths, cols)]
        headers = [
            _align_header(h, a, minw) for h, a, minw in zip(headers, aligns, minwidths)
        ]
    else:
        minwidths = [width_fn(c[0]) for c in cols]
    rows = list(zip(*cols))

    if not isinstance(tablefmt, TableFormat):
        tablefmt = _table_formats.get(tablefmt, _table_formats["orgmode"])

    # make sure values don't have newlines or tabs in them, except for tsv output_format where
    # we have to add spaces in order to simulate tab separators
    rows = [
        [
            str(c)
            .replace("\n", "")
            .replace("\t", " " if is_tsv else "")
            .replace("\r", "")
            for c in r
        ]
        for r in rows
    ]

    return _format_table(tablefmt, headers, rows, minwidths, aligns)


if __name__ == "__main__":
    data = [
        ("gene_symbols", ["CPB1"]),
        ("clinical_significance", "other"),
        ("clinical_origin", ["somatic"]),
        ("alternate_alleles", ["T"]),
    ]
    print(
        tabulate(data, headers=("Fields", "Data"), aligns=("right", "left"), sort=True)
    )
    print(
        tabulate(data, headers=("Fields", "Data"), aligns=("right", "left"), sort=False)
    )



"""QuartzBio Task API Resource"""

from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource


class Task(ListableAPIResource, UpdateableAPIResource):
    """
    Tasks are operations on datasets or vaults.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("task_display_name", "Task Type"),
        ("task_id", "Task ID"),
        ("description", "Description"),
        ("status", "Status"),
        ("created_at", "Created"),
    )

    SLEEP_WAIT_DEFAULT = 5.0

    @property
    def child_object(self):
        """Get Task child object class"""
        from . import types

        child_klass = types.get(self.task_type.split(".")[1])
        return child_klass.retrieve(self.task_id, client=self._client)

    def follow(self, sleep_seconds=SLEEP_WAIT_DEFAULT):
        """Follow the child object but do not loop"""
        self.child_object.follow(loop=False, sleep_seconds=sleep_seconds)

    def cancel(self):
        """Cancel a task"""
        _status = self.status
        self.status = "canceled"
        try:
            self.save()
        except:
            # Reset status to what it was before
            # status update failure
            self.status = _status
            raise



# -*- coding: utf-8 -*-

from .client import client

import logging

logger = logging.getLogger("quartzbio")


class Annotator(object):
    """
    Runs the synchronous annotate endpoint against
    batches of results from a query.
    """

    CHUNK_SIZE = 100

    # Allows pre-setting a QuartzBioClient
    _client = None

    def __init__(self, fields, **kwargs):
        self._client = kwargs.pop("client", None) or self._client or client

        self.buffer = []
        self.fields = fields

        # Pop annotator_params from kwargs
        annotator_param_keys = [
            "annotator",
            "debug",
            "include_errors",
            "post_annotation_expression",
            "pre_annotation_expression",
        ]
        self.annotator_params = {}
        for key in annotator_param_keys:
            if key in kwargs:
                self.annotator_params[key] = kwargs.pop(key)

        self.data = kwargs.get("data")

    def annotate(self, records, **kwargs):
        """Annotate a set of records with stored fields.

        Args:
            records: A list or iterator (can be a Query object)
            chunk_size: The number of records to annotate at once (max 500).

        Returns:
            A generator that yields one annotated record at a time.
        """
        # Update annotator_params with any kwargs
        self.annotator_params.update(**kwargs)
        chunk_size = self.annotator_params.get("chunk_size", self.CHUNK_SIZE)

        chunk = []
        for i, record in enumerate(records):
            chunk.append(record)
            if (i + 1) % chunk_size == 0:
                for r in self._execute(chunk):
                    yield r
                chunk = []

        if chunk:
            for r in self._execute(chunk):
                yield r
            chunk = []

    def _execute(self, chunk):
        data = {
            "records": chunk,
            "fields": self.fields,
            "annotator_params": self.annotator_params,
            "data": self.data,
        }

        for r in self._client.post("/v1/annotate", data)["results"]:
            yield r


class Expression(object):
    """Runs a single QuartzBio expression."""

    # Allows pre-setting a QuartzBioClient
    _client = None

    def __init__(self, expr, **kwargs):
        self.expr = expr
        self._client = kwargs.get("client") or self._client or client

    def evaluate(self, data=None, data_type="string", is_list=False):
        """Evaluates the expression with the provided context and format."""
        payload = {
            "data": data,
            "expression": self.expr,
            "data_type": data_type,
            "is_list": is_list,
        }
        res = self._client.post("/v1/evaluate", payload)
        return res["result"]

    def __repr__(self):
        return '<Expression "{0}">'.format(self.expr)



from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource


class Beacon(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    UpdateableAPIResource,
):
    """
    Beacons provide entity-based search endpoints for datasets. Beacons
    must be created within Beacon Sets.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("title", "Title"),
        ("description", "Description"),
        ("vault_object_id", "Object ID"),
    )

    def _query_url(self):
        if "query_url" not in self:
            if "id" not in self or not self["id"]:
                raise Exception(
                    "No Beacon ID was provided. "
                    "Please instantiate the Beacon"
                    "object with an ID."
                )
            return self.instance_url() + "/query"
        return self["query_url"]

    def query(self, query, entity_type=None):
        data = {"query": query, "entity_type": entity_type}
        return self._client.post(self._query_url(), data)



from .apiresource import ListableAPIResource
from .apiresource import DeletableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import CreateableAPIResource
from .quartzbio_object import convert_to_quartzbio_object


class Group(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    DeletableAPIResource,
):
    """
    A Group represents a group of users with shared permissions for a vault.
    """

    RESOURCE_VERSION = 1

    LIST_FIELDS = (
        ("id", "ID"),
        ("name", "Name"),
        ("memberships_count", "Members"),
        ("vaults_count", "Vaults"),
        ("role", "Role"),
        ("description", "Description"),
    )

    def members(self, **params):
        response = self._client.get(self.memberships_url, params)
        results = convert_to_quartzbio_object(response, client=self._client)
        results.set_tabulate(
            ["id", "full_name", "username", "email", "role"],
            headers=["ID", "Full Name", "Username", "Email", "Role"],
            aligns=["left", "left", "left", "left", "left"],
            sort=False,
        )

        return results

    def _get_vaults(self, **params):
        response = self._client.get(self.vaults_url, params)
        return convert_to_quartzbio_object(response, client=self._client)

    def vaults(self, **params):
        results = self._get_vaults(**params)
        results.set_tabulate(
            ["id", "name", "permission", "vault_permissions"],
            headers=["ID", "Vault", "Permission", "Permissions Detail"],
            aligns=["left", "left", "left", "left"],
            sort=False,
        )

        return results

    def datasets(self, **params):
        from . import Object

        vaults = self._get_vaults(**params)
        objects_url = (
            Object.class_url()
            + "?object_type=dataset&"
            + "&".join(["vault_id={0}".format(v.id) for v in vaults])
        )
        response = self._client.get(objects_url, params)
        datasets = convert_to_quartzbio_object(response, client=self._client)
        datasets.set_tabulate(
            [
                "id",
                "vault_name",
                "path",
                "dataset_documents_count",
                "dataset_description",
            ],
            headers=["ID", "Vault", "Path", "Documents", "Description"],
            aligns=["left", "left", "left", "left", "left"],
            sort=False,
        )

        return datasets



# Add common quartzbio classes and methods our namespace here so that
# inside the ipython shell users don't have run imports
import quartzbio  # noqa
from quartzbio import login  # noqa
from quartzbio import Annotator  # noqa
from quartzbio import Application  # noqa
from quartzbio import Beacon  # noqa
from quartzbio import BeaconSet  # noqa
from quartzbio import BatchQuery  # noqa
from quartzbio import Dataset  # noqa
from quartzbio import DatasetCommit  # noqa
from quartzbio import DatasetExport  # noqa
from quartzbio import DatasetField  # noqa
from quartzbio import DatasetImport  # noqa
from quartzbio import DatasetMigration  # noqa
from quartzbio import DatasetTemplate  # noqa
from quartzbio import Expression  # noqa
from quartzbio import Filter  # noqa
from quartzbio import GenomicFilter  # noqa
from quartzbio import Manifest  # noqa
from quartzbio import Object  # noqa
from quartzbio import Query  # noqa
from quartzbio import QuartzBioClient  # noqa
from quartzbio import QuartzBioError  # noqa
from quartzbio import User  # noqa
from quartzbio import Vault  # noqa
from quartzbio import Task  # noqa
from quartzbio import VaultSyncTask  # noqa
from quartzbio import ObjectCopyTask  # noqa
from quartzbio import SavedQuery  # noqa
from quartzbio import DatasetRestoreTask  # noqa
from quartzbio import DatasetSnapshotTask  # noqa
from quartzbio import GlobalSearch  # noqa
from quartzbio import Group  # noqa
from quartzbio.utils.printing import pager  # noqa

# Add some convenience functions to the interactive shell
from quartzbio.cli.auth import logout  # noqa
from quartzbio.cli.auth import whoami  # noqa
from quartzbio.cli.auth import get_credentials  # noqa

whoami()




import dash
import flask

import os
from random import randint

import quartzbio

from .quartzbio_auth import QuartzBioAuth


class QuartzBioDash(dash.Dash):
    APP_URL = os.environ.get("APP_URL", "http://127.0.0.1:8050")
    QUARTZBIO_URL = os.environ.get("QUARTZBIO_URL", "http://my.quartzbio.com")
    SECRET_KEY = os.environ.get("SECRET_KEY", str(randint(0, 1000000)))

    def __init__(self, name, *args, **kwargs):
        # Default HTML page title
        self.title = kwargs.pop("title", name)

        app_url = kwargs.pop("app_url", self.APP_URL)
        quartzbio_url = kwargs.pop("quartzbio_url", self.QUARTZBIO_URL)
        api_host = kwargs.pop("api_host", None)  # Extract api_host before passing to Dash

        # OAuth2 credentials
        client_id = kwargs.pop("client_id", os.environ.get("CLIENT_ID"))
        client_secret = kwargs.pop("client_secret", os.environ.get("CLIENT_SECRET"))
        grant_type = kwargs.pop("grant_type", None)
        salt = kwargs.pop("salt", None) or name

        server = flask.Flask(name)
        server.secret_key = kwargs.pop("secret_key", self.SECRET_KEY)
        kwargs["server"] = server

        super(QuartzBioDash, self).__init__(name, *args, **kwargs)

        if client_id:
            self.auth = QuartzBioAuth(
                self,
                app_url,
                client_id,
                salt=salt,
                client_secret=client_secret,
                grant_type=grant_type,
                quartzbio_url=quartzbio_url,
                api_host=api_host,
            )
        else:
            self.auth = None
            print(
                "WARNING: No QuartzBio client ID found. "
                "Your app (but not your data) will be publicly accessible."
            )

        @server.before_request
        def set_quartzbio_client():
            oauth_token = flask.request.cookies.get(QuartzBioAuth.TOKEN_COOKIE_NAME)
            if oauth_token:
                flask.g.client = quartzbio.QuartzBioClient(
                    token=oauth_token, token_type="Bearer"
                )
            else:
                # Use global credentials (if any)
                flask.g.client = quartzbio.QuartzBioClient()



# -*- coding: utf-8 -*-

from abc import ABCMeta
import json
import uuid

from .client import client
from .utils.printing import pretty_int
from .utils.tabulate import tabulate
from .errors import QuartzBioError

import copy
import logging

logger = logging.getLogger("quartzbio")


class Filter(object):
    """
    Filter objects.

    Makes it easier to create filters cumulatively using ``&`` (and),
    ``|`` (or) and ``~`` (not) operations.

    For example::

        f = Filter()
        f &= Filter(price='Free')
        f |= Filter(style='Mexican')

    creates a filter "price = 'Free' or style = 'Mexican'".

    Each set of kwargs in a `Filter` are ANDed together:

      * `<field>='<value>'` matches if the field is that exact value
      * `<field>__in=[<item1>, ...]` matches any of the terms <item1> and so on
      * `<field>__range=[<start>, <end>]` matches anything from <start>
         to <end>
      * `<field>__between=[<start>, <end>]` matches anything between <start> to
         <end> not include either <start> or <end>

    String terms are not analyzed and are always assumed to be exact matches.

    Numeric columns can be selected by range using:

        * `<field>__gt`: greater than
        * `<field>__gte`: greater than or equal to
        * `<field>__lt`: less than
        * `<field>__lte`: less than or equal to

    Field action examples:

        dataset.query(gene__in=['BRCA', 'GATA3'],
                      chr='3',
                      start__gt=10000,
                      end__lte=20000)
    """

    def __init__(self, *raw_filters, **filters):
        """Creates a Filter"""
        # Set deepcopy to False for faster Filter building
        self.deepcopy = True

        filters = list(filters.items())
        for flt in raw_filters:
            try:
                flt = json.loads(flt)
                # If the result is a dict, wrap in a list.
                if isinstance(flt, dict):
                    flt = [flt]
            except:
                raise Exception(
                    "Invalid raw filter, must be a JSON string: {}".format(flt)
                )

            filters += flt

        if len(filters) > 1:
            self.filters = [{"and": filters}]
        else:
            self.filters = filters

    def __repr__(self):
        return "<Filter {0}>".format(self.filters)

    def _combine(self, other, conn="and"):
        """
        OR and AND will create a new Filter, with the filters from both Filter
        objects combined with the connector `conn`.
        """
        f = Filter()
        f.deepcopy = self.deepcopy and other.filters

        if f.deepcopy:
            self_filters = copy.deepcopy(self.filters)
            other_filters = copy.deepcopy(other.filters)
        else:
            self_filters = self.filters
            other_filters = other.filters

        if not self.filters:
            f.filters = other_filters
        elif not other.filters:
            f.filters = self_filters
        elif conn in self.filters[0]:
            f.filters = self_filters
            f.filters[0][conn].extend(other_filters)
        elif conn in other.filters[0]:
            f.filters = other_filters
            f.filters[0][conn].extend(self_filters)
        else:
            f.filters = [{conn: self_filters + other_filters}]

        return f

    def __or__(self, other):
        return self._combine(other, "or")

    def __and__(self, other):
        return self._combine(other, "and")

    def __invert__(self):
        f = Filter()
        f.deepcopy = self.deepcopy

        if f.deepcopy:
            self_filters = copy.deepcopy(self.filters)
        else:
            self_filters = self.filters

        if len(self.filters) == 0:
            # no change
            f.filters = []
        elif (
            len(self.filters) == 1
            and isinstance(self.filters[0], dict)
            and self.filters[0].get("not", {})
        ):
            # if the filters are already a single dictionary containing a 'not'
            # then swap out the 'not'
            f.filters = [self_filters[0]["not"]]
        else:
            # length of self.filters should never be more than 1
            # 'not' blocks can contain only dicts or a single tuple filter
            # so we get the first element from the filter list
            f.filters = [{"not": self_filters[0]}]

        return f


class GenomicFilter(Filter):
    """
    Helper class that generates filters on genomic coordinates.

    Range filtering only works on "genomic" datasets
    (where dataset.is_genomic is True).
    """

    # Standardized fields for genomic coordinates in QuartzBio
    FIELD_START = "genomic_coordinates.start"
    FIELD_STOP = "genomic_coordinates.stop"
    FIELD_CHR = "genomic_coordinates.chromosome"

    @classmethod
    def from_string(cls, string, exact=False):
        """
        Handles UCSC-style range queries (chr1:100-200)
        """
        try:
            chromosome, pos = string.split(":")
        except ValueError:
            raise ValueError('Please use UCSC-style format: "chr2:1000-2000"')

        if "-" in pos:
            start, stop = pos.replace(",", "").split("-")
        else:
            start = stop = pos.replace(",", "")

        return cls(chromosome, start, stop, exact=exact)

    def __init__(self, chromosome, start, stop=None, exact=False):
        """
        This class supports single position and range filters.

        By default, the filter will match any record that overlaps with
        the position or range specified. Exact matches must be explicitly
        specified using the `exact` parameter.
        """
        try:
            # Allows start=None to filter items with no position
            if start is not None:
                start = int(start)

            if stop is None:
                stop = start
            else:
                stop = int(stop)
        except ValueError:
            raise ValueError("Start and stop positions must be integers (or None)")

        if exact or start is None:
            # Handle the case where start is None because sometimes
            # a record will have only the chromosome set (no positions).
            f = Filter(**{self.FIELD_START: start, self.FIELD_STOP: stop})
        else:
            f = Filter(
                **{self.FIELD_START + "__lte": start, self.FIELD_STOP + "__gte": stop}
            )
            if start != stop:
                f = f | Filter(**{self.FIELD_START + "__range": [start, stop]})
                f = f | Filter(**{self.FIELD_STOP + "__range": [start, stop]})

        if chromosome is None:
            f = f & Filter(**{self.FIELD_CHR: None})
        else:
            f = f & Filter(**{self.FIELD_CHR: str(chromosome).replace("chr", "")})
        self.filters = f.filters

    def __repr__(self):
        return "<GenomicFilter {0}>".format(self.filters)


class QueryBase(object, metaclass=ABCMeta):
    """
    A helper abstract mixin class that contains
    common methods for Query and QueryFile classes.
    """

    # The default number of records fetched in one go.
    DEFAULT_PAGE_SIZE = 100

    # The maximum number of records fetched in one go.
    MAX_PAGE_SIZE = 10 * 1000

    # INF represents an integer version of 'float('inf')'
    # because it could not be typecasted to integer
    INF = 10**15

    # Special case for Query/QueryFile class to pre-set QuartzBioClient
    _client = None

    def limit(self, limit):
        """
        Returns a new Query/QueryFile instance with the new
        limit values.
        """
        return self._clone(limit=limit)

    def count(self):
        """
        Returns the total number of results returned by a query.
        The count is dependent on the filters, but independent of any limit.
        It is like SQL:
        SELECT COUNT(*) FROM <table> [WHERE condition].
        See also __len__ for a function that is dependent on limit.

        """
        # self.total will warm up the response if it needs to
        return self.total

    def __len__(self):
        """
        Returns the total number of results returned in a query. It is the
        number of items you can iterate over.

        In contrast to count(), the result does take into account any limit
        given. In SQL it is like:

              SELECT COUNT(*) FROM (
                 SELECT * FROM <table> [WHERE condition] [LIMIT number]
              )
        """
        return (
            min(self._limit, self.count()) if not self.count() is None else self._limit
        )

    def __nonzero__(self):
        return bool(len(self))

    @property
    def _buffer(self):
        if self._response is None:
            logger.debug("warmup (buffer)")
            self.execute(self._slice.start if self._slice else 0)
        return self._response["results"]

    def __repr__(self):
        # Check that Query/QueryFile object does not have any previous errors
        # otherwise, raise the error.
        total_count = 0
        try:
            total_count = len(self)
        except TypeError:
            total_count = float("inf")

        if self._error:
            raise self._error

        if total_count == 0:
            return "Query returned 0 results."

        placeholder = "many more results (total unknown)"

        if getattr(self, "_is_join", False):
            return "\n%s\n\n... %s." % (
                tabulate(
                    list(self._buffer[0].items()),
                    ["Fields", "Data"],
                    aligns=["right", "left"],
                    sort=True,
                ),
                placeholder,
            )
        elif not hasattr(self, "_output_format"):
            return "\n%s\n\n... %s more results." % (
                tabulate(
                    list(self._buffer[0].items()),
                    ["Fields", "Data"],
                    aligns=["right", "left"],
                    sort=True,
                ),
                pretty_int(total_count - 1),
            )
        else:
            is_tsv = self._output_format == "tsv"

            # this is the only case when we know the exact number of total records
            if total_count < self._limit:
                placeholder = "{} more results".format(
                    pretty_int(max(total_count - 9, 0))
                )

            return "\n%s\n\n... %s." % (
                tabulate(
                    list(enumerate(self._buffer[:10])),
                    ["Row", "Data"],
                    aligns=["right", "left"],
                    is_tsv=is_tsv,
                ),
                placeholder,
            )

    def __getattr__(self, key):
        if self._response is None:
            logger.debug("warmup (__getattr__: %s)" % key)
            self.execute(self._slice.start if self._slice else 0)

        # Check that Query/QueryFile object does not have any previous errors
        # otherwise, raise the error.
        # execute() sets the error, so the check is placed after it.
        if self._error:
            raise self._error

        if key in self._response:
            return self._response[key]

        raise AttributeError(
            "'%s' object has no attribute '%s'" % (self.__class__.__name__, key)
        )

    @staticmethod
    def bounded_slice(_slice):
        return slice(
            _slice.start if _slice.start is not None else 0,
            _slice.stop if _slice.stop is not None else float("inf"),
        )

    @staticmethod
    def as_slice(slice_or_idx):
        if isinstance(slice_or_idx, slice):
            return QueryBase.bounded_slice(slice_or_idx)
        return slice(slice_or_idx, slice_or_idx + 1)

    def __getitem__(self, key):
        """
        Retrieve an item or slice from the result set.

        Query/QueryFile's do not support negative indexing.

        :Parameters:
        - `key`: The requested slice range or index.
        """
        if not isinstance(key, (slice, int)):
            raise TypeError

        if isinstance(key, slice):
            key = QueryBase.bounded_slice(key)
            start = 0 if key.start is None else key.start
            stop = float("inf") if key.stop is None else key.stop

            if start < 0 or stop < 0 or start > stop:
                raise ValueError("Negative indexing is not supported")

            # If a slice is already set, the new slice should be relative
            if self._slice:
                start += self._slice.start
                stop = min(self._slice.start + stop, self._slice.stop)
                # Make sure the new relative start position is within
                # the previous slice.
                if start >= self._slice.stop:
                    return []

            # We need to make a few requests to get the data.
            # We should respect the user's limit if it is smaller than slice.
            # To prevent the state of page_size/offset from being stored,
            # we'll clone this object first.
            q = self._clone()
            q._limit = min(stop - start, self._limit)
            # Setting slice will signal to the iter methods the page_offset.
            q._slice = slice(start, stop)
            return q

        # Not a slice (key is an int)
        if key < 0:
            raise ValueError("Negative indexing is not supported")

        # If a slice already exists, the key is relative to that slice
        if self._slice:
            key = key + self._slice.start
            if key >= self._slice.stop:
                raise IndexError("list index out of range")

        # Use key as the new page_offset and fetch a new page of results
        q = self._clone()
        q._limit = min(1, self._limit)  # Limit may be 0
        q.execute(key)
        return q._buffer[0]

    def __iter__(self):
        # e.g. [r for r in results] will NOT call __getitem__ and
        # requires that we start iteration from the 0th element
        self.execute(self._slice.start if self._slice else 0)

        # Reset the cursor
        self._cursor = 0  # Count the number of results returned
        self._buffer_idx = 0  # The current position within the buffer

        return self

    def __next__(self):
        """Python 3"""
        return self.next()

    def next(self):
        """
        Allows the Query/QueryFile object to be an iterable.

        This method will iterate through a cached result set
        and fetch successive pages as required.

        A `StopIteration` exception will be raised when there aren't
        any more results available or when the requested result
        slice range or limit has been fetched.

        Returns: The next result.
        """
        if not hasattr(self, "_cursor"):
            # Iterator not initialized yet
            self.__iter__()

        # Check if a current object is the join query
        _is_join = getattr(self, "_is_join", False)

        # len(self) returns `min(limit, total)` results
        try:
            if not _is_join and self._cursor == len(self):
                raise StopIteration
        except TypeError:
            # len(self) is unknown so just continue normally
            pass

        if self._buffer_idx == len(self._buffer):
            if _is_join:
                if self._next_offset >= self._limit:
                    # Since joins can return more results than we expect (due to `explode`)
                    # manually ensure that we haven't gone above the requested limit (default inf)
                    raise StopIteration
                self.execute(self._next_offset)
            else:
                self.execute(self._page_offset + self._buffer_idx)
            self._buffer_idx = 0

        if not self._buffer:
            raise StopIteration

        self._cursor += 1
        self._buffer_idx += 1

        return self._buffer[self._buffer_idx - 1]

    def filter(self, *filters, **kwargs):
        """
        Returns this Query/QueryFile instance with the query args combined with
        existing set with AND.

        kwargs are simply passed to a new Filter object and combined to any
        other filters with AND.

        By default, everything is combined using AND. If you provide
        multiple filters in a single filter call, those are ANDed
        together. If you provide multiple filters in multiple filter
        calls, those are ANDed together.

        If you want something different, use the F class which supports
        ``&`` (and), ``|`` (or) and ``~`` (not) operators. Then call
        filter once with the resulting Filter instance.
        """
        f = list(filters)

        if kwargs:
            f += [Filter(**kwargs)]

        return self._clone(filters=f)

    @classmethod
    def _process_filters(cls, filters):
        """Takes a list of filters and returns JSON

        :Parameters:
        - `filters`: List of Filters, (key, val) tuples, or dicts

        Returns: List of JSON API filters
        """
        data = []

        # Filters should always be a list
        for f in filters:
            if isinstance(f, Filter):
                if f.filters:
                    data.extend(cls._process_filters(f.filters))
            elif isinstance(f, dict):
                key = list(f.keys())[0]
                val = f[key]

                if isinstance(val, dict):
                    # pass val (a dict) as list
                    # so that it gets processed properly
                    filter_filters = cls._process_filters([val])
                    if len(filter_filters) == 1:
                        filter_filters = filter_filters[0]
                    data.append({key: filter_filters})
                else:
                    data.append({key: cls._process_filters(val)})
            else:
                data.extend((f,))

        return data


class Query(QueryBase):
    """
    A Query API request wrapper that generates a request from Filter objects,
    and can iterate through streaming result sets.
    """

    def __init__(
        self,
        dataset_id,
        query=None,
        genome_build=None,
        filters=None,
        fields=None,
        exclude_fields=None,
        entities=None,
        ordering=None,
        limit=float("inf"),
        page_size=QueryBase.DEFAULT_PAGE_SIZE,
        result_class=dict,
        target_fields=None,
        annotator_params=None,
        debug=False,
        **kwargs,
    ):
        """
        Creates a new Query object.

        :Parameters:
          - `dataset_id`: Unique ID of dataset to query.
          - `query` (optional): An optional query string.
          - `genome_build`: The genome build to use for the query.
          - `filters` (optional): Filter or List of filter objects.
          - `fields` (optional): List of specific fields to retrieve.
          - `exclude_fields` (optional): List of specific fields to exclude.
          - `entities` (optional): List of entity tuples to filter on.
          - `ordering` (optional): List of fields to order the results by.
          - `limit` (optional): Maximum number of query results to return.
          - `page_size` (optional): Number of results to fetch per query page.
          - `result_class` (optional): Class of object returned by query.
          - `target_fields` (optional): Add target fields to annotate the query results.
          - `annotator_params` (optional): For use with `target_fields` to adjust annotator parameters.
          - `debug` (optional): Sends debug information to the API.
        """
        self._dataset_id = dataset_id
        self._data_url = "/v2/datasets/{0}/data".format(dataset_id)
        self._query = query
        self._genome_build = genome_build
        self._fields = fields
        self._exclude_fields = exclude_fields
        self._entities = entities
        self._ordering = ordering
        self._result_class = result_class
        self._target_fields = target_fields
        self._annotator_params = annotator_params
        self._debug = debug
        self._error = None
        self._is_join = False

        if filters:
            if isinstance(filters, Filter):
                filters = [filters]
        else:
            filters = []
        self._filters = filters

        # init response and cursor
        self._response = None
        # Limit defines the total number of results that will be returned
        # from a query involving 1 or more pagination requests.
        self._limit = limit
        # Page size/offset are the low level API limit and offset params.
        self._page_size = int(page_size)
        # Page offset can only be set by execute(). It is always set to the
        # current absolute offset contained in the buffer.
        self._page_offset = None
        # slice is set when the Query is being sliced.
        # In this case, __iter__() and next() will not
        # reset the page_offset to 0 before iterating.
        self._slice = None

        # parameter error checking
        if self._limit < 0:
            raise Exception("'limit' parameter must be >= 0")

        if not 0 < self._page_size <= self.MAX_PAGE_SIZE:
            raise Exception(
                "'page_size' parameter must be in " "range [1, {}]".format(
                    self.MAX_PAGE_SIZE
                )
            )

        # Set up the QuartzBioClient
        # (kwargs overrides pre-set, which overrides global)
        self._client = kwargs.get("client") or self._client or client

    def _clone(self, filters=None, limit=None):
        new = self.__class__(
            self._dataset_id,
            query=self._query,
            genome_build=self._genome_build,
            limit=self._limit,
            fields=self._fields,
            exclude_fields=self._exclude_fields,
            entities=self._entities,
            ordering=self._ordering,
            page_size=self._page_size,
            result_class=self._result_class,
            target_fields=self._target_fields,
            annotator_params=self._annotator_params,
            debug=self._debug,
            client=self._client,
        )
        new._filters += self._filters

        if filters:
            new._filters += filters

        if limit is not None:
            new._limit = limit

        return new

    def range(self, chromosome, start, stop, exact=False):
        """
        Shortcut to do range filters on genomic datasets.
        """
        return self._clone(filters=[GenomicFilter(chromosome, start, stop, exact)])

    def position(self, chromosome, position, exact=False):
        """
        Shortcut to do a single position filter on genomic datasets.
        """
        return self._clone(filters=[GenomicFilter(chromosome, position, exact=exact)])

    def facets(self, *args, **kwargs):
        """
        Returns a dictionary with the requested facets.

        The facets function supports string args, and keyword
        args.

        q.facets('field_1', 'field_2') will return facets for
        field_1 and field_2.
        q.facets(field_1={'limit': 0}, field_2={'limit': 10})
        will return all facets for field_1 and 10 facets for field_2.
        """
        # Combine args and kwargs into facet format.
        facets = dict((a, {}) for a in args)
        facets.update(kwargs)

        if not facets:
            raise AttributeError("Faceting requires at least one field")

        for f in facets.keys():
            if not isinstance(f, str):
                raise AttributeError("Facet field arguments must be strings")

        q = self._clone()
        q._limit = 0
        q.execute(offset=0, facets=facets)
        return q._response.get("facets")

    def __len__(self):
        """
        Returns the total number of results returned in a query. It is the
        number of items you can iterate over.

        In contrast to count(), the result does take into account any limit
        given. In SQL it is like:

              SELECT COUNT(*) FROM (
                 SELECT * FROM <table> [WHERE condition] [LIMIT number]
              )
        """
        if self._is_join:
            return len(self._buffer)

        return super(Query, self).__len__()

    def _build_query(self, **kwargs):
        q = {}

        if self._query:
            q["query"] = self._query

        if self._filters:
            filters = self._process_filters(self._filters)
            if len(filters) > 1:
                q["filters"] = [{"and": filters}]
            else:
                q["filters"] = filters

        if self._fields is not None:
            q["fields"] = self._fields

        if self._exclude_fields is not None:
            q["exclude_fields"] = self._exclude_fields

        if self._entities is not None:
            q["entities"] = self._entities

        if self._ordering is not None:
            q["ordering"] = self._ordering

        if self._genome_build is not None:
            q["genome_build"] = self._genome_build

        if self._debug:
            q["debug"] = "True"

        if self._target_fields:
            q["target_fields"] = self._target_fields

        if self._annotator_params:
            q["annotator_params"] = self._annotator_params

        # Add or modify query parameters
        # (used by BatchQuery and facets)
        q.update(**kwargs)

        return q

    def execute(self, offset=0, **query):
        """
        Executes a query. Additional query parameters can be passed
        as keyword arguments.

        Returns: The request parameters and the raw query response.
        """
        _params = self._build_query(**query)
        self._page_offset = offset

        _params.update(
            offset=self._page_offset, limit=min(self._page_size, self._limit)
        )

        if self._is_join:
            # We do not know the exact total number of records in join because it
            # is dynamically calculated in internal expression in target_fields in
            # join() method, therefore we have to change limit in the last
            # subsequent request in order to get the given number of records from query_a
            _params["limit"] = min(
                self._page_size, abs(self._limit - self._page_offset)
            )
            self._next_offset = self._page_offset + min(self._page_size, self._limit)

        logger.debug(
            "executing query. from/limit: %6d/%d"
            % (_params["offset"], _params["limit"])
        )

        # If the request results in a QuartzBioError (ie bad filter) set the error.
        try:
            self._response = self._client.post(self._data_url, _params)
        except QuartzBioError as e:
            self._error = e
            raise

        logger.debug(
            "query response took: %(took)d ms, total: %(total)d" % self._response
        )
        return _params, self._response

    def fields(self):
        """Returns all expected fields that will be found in the results."""

        from quartzbio import Dataset

        fields = [f for f in Dataset(self._dataset_id, client=self._client).fields()]
        if self._fields:
            fields = [f for f in fields if f.name in self._fields]
        if self._exclude_fields:
            fields = [f for f in fields if f.name not in self._exclude_fields]

        return fields

    def export(self, format="json", follow=True, limit=None, **kwargs):
        from quartzbio import DatasetExport

        params = self._build_query()
        params.pop("offset", None)
        params.pop("ordering", None)

        # if limit is not set use max limit
        if not limit and self._limit < float("inf"):
            limit = self._limit

        if limit:
            params["limit"] = limit

        # target_fields and annotator_params must be passed
        # directly into the migration.
        # Prefer explicitly passed-in values before query values.
        target_fields = kwargs.pop("target_fields", None) or params.pop(
            "target_fields", None
        )
        annotator_params = kwargs.pop("annotator_params", {}) or params.pop(
            "annotator_params", {}
        )

        export = DatasetExport.create(
            dataset_id=self._dataset_id,
            format=format,
            params=params,
            target_fields=target_fields,
            annotator_params=annotator_params,
            client=self._client,
            **kwargs,
        )

        if follow:
            export.follow()

        return export

    def migrate(self, target, follow=True, **kwargs):
        """
        Migrate the data from the Query to a target dataset.

        Valid optional kwargs include:

        * target_fields
        * include_errors
        * validation_params
        * metadata
        * commit_mode

        """
        from quartzbio import DatasetMigration

        # Target can be provided as an object or an ID.
        try:
            target_id = target.id
        except AttributeError:
            target_id = target

        # If a limit is set in the Query and not overridden here, use it.
        limit = kwargs.pop("limit", None)
        if not limit and self._limit < float("inf"):
            limit = self._limit

        # Build the source_params
        params = self._build_query(limit=limit)
        params.pop("offset", None)
        params.pop("ordering", None)

        # target_fields and annotator_params must be passed
        # directly into the migration.
        # Prefer explicitly passed-in values before query values.
        target_fields = kwargs.pop("target_fields", None) or params.pop(
            "target_fields", None
        )
        annotator_params = kwargs.pop("annotator_params", {}) or params.pop(
            "annotator_params", {}
        )

        migration_resp = DatasetMigration.create(
            source_id=self._dataset_id,
            target_id=target_id,
            source_params=params,
            target_fields=target_fields,
            annotator_params=annotator_params,
            client=self._client,
            **kwargs,
        )

        if follow:
            # If migration was created with the parallel flag, then multiple
            # migration objects will be returned.
            if "data" in migration_resp:
                for migration in migration_resp["data"]:
                    migration.follow()
            else:
                migration_resp.follow()

        return migration_resp

    def annotate(self, fields, **kwargs):
        from quartzbio.annotate import Annotator

        return Annotator(fields, client=self._client, **kwargs).annotate(self)

    def join(self, query_b, key, key_b=None, prefix="b_", always_prefix=False):
        """Performs a left outer join between the current
        query (query A) and another query (query B).

        Set prefix to None to use a random prefix.
        Enable always_prefix to always apply a prefix.
        """

        # Generate a random ID for the transient query field
        join_id = str(uuid.uuid4())[:8]
        # If prefix is cleared, use the join ID
        if not prefix:
            prefix = join_id + "_"

        # If no key_b is provided, use the same key as for A
        key_b = key_b or key

        # Prepare new returned query object
        new_query = self._clone()

        # Initialize _explode_fields attribute if it does not exist
        new_query._explode_fields = getattr(self, "_explode_fields", None) or []

        # Set list of existing field names to avoid overwriting fields
        # in the join.
        existing_field_names = [f.name for f in self.fields()]
        if new_query._target_fields:
            existing_field_names += [f["name"] for f in new_query._target_fields]
        else:
            new_query._target_fields = []

        # Prepare the filters for the B query expression
        query_params = query_b._build_query()
        base_filter = '["{}", get(record, "{}")]'.format(key_b, key)
        if query_params.get("filters"):
            filters = '[{{"and": [{}, {}]}}]'.format(
                base_filter, query_params["filters"][0]
            )
        else:
            filters = "[{}]".format(base_filter)

        # Try to use a unique field for the sub-query data
        query_b_fields = query_b.fields()

        # If a joining key in both datasets is the same and
        # it is not the only one field in a query_b then remove it from query_b
        if key_b == key and not (
            len(query_b_fields) == 1 and query_b_fields[0].name == key_b
        ):
            query_b_fields = [item for item in query_b_fields if not item.name == key_b]

        query_b_join_field_name = "join_{}".format(join_id)
        target_fields = [
            {
                "name": query_b_join_field_name,
                "data_type": "object",
                "is_list": True,
                "is_transient": True,
                "expression": """
                    dataset_query(
                        "{}",
                        fields={},
                        filters={},
                        entities={},
                        limit={},
                        ordering={}
                    ) if get(record, "{}") else {{}}
                """.format(
                    query_b._dataset_id,
                    [f.name for f in query_b_fields] + [key_b],
                    filters,
                    query_params.get("entities"),
                    query_b._limit if isinstance(query_b._limit, int) else 100000,
                    query_b._ordering,
                    key,
                ),
            }
        ]

        # Get list of fields to join from query B
        for field in query_b_fields:
            # If "always prefix" is enable, or the field name is found
            # in existing list of names, add the prefix.
            if always_prefix or field.name in existing_field_names:
                name = prefix + field.name
            else:
                name = field.name

            if name in existing_field_names:
                raise Exception(
                    "Field '{}' found in both queries, "
                    "please use a different prefix.".format(name)
                )

            # Add a newly created field to list that will be passed to the explode function
            new_query._explode_fields.append(name)

            # Extract all the field values from each joined record
            expression = '[get(item, "{}") for item in get(record, "{}")]'.format(
                field.name, query_b_join_field_name
            )

            if field.is_list:
                # Handle case where the sub-query contains a list of lists of <field.data_type>,
                # which will happen if the field contains a list of values.
                # In this case, we will flatten it to avoid returning a list of lists.
                expression = (
                    "[item for sublist in " + expression + " for item in sublist]"
                )

            target_fields.append(
                {
                    "name": name,
                    "title": field.title,
                    "data_type": field.data_type,
                    "ordering": field.ordering,
                    "is_list": True,
                    "is_transient": False,
                    "expression": expression,
                    "depends_on": [query_b_join_field_name],
                }
            )

        # Add to any existing target fields
        new_query._target_fields += target_fields

        # Preserve existing annotator params (pre_annotation_expression only)
        if not new_query._annotator_params:
            new_query._annotator_params = {}

        new_query._annotator_params["post_annotation_expression"] = (
            "explode(record, fields={})".format(new_query._explode_fields)
        )

        new_query._is_join = True
        return new_query


class BatchQuery(object):
    """
    BatchQuery accepts a list of Query objects and executes them
    in a single request to /v2/batch_query.
    """

    # Allows pre-setting a QuartzBioClient
    _client = None

    def __init__(self, queries, **kwargs):
        """
        Expects a list of Query objects.
        """
        if not isinstance(queries, list):
            queries = [queries]

        self._queries = queries
        self._client = kwargs.get("client") or self._client or client

    def _build_query(self):
        query = {"queries": []}

        for i in self._queries:
            _params = i._build_query(
                dataset=i._dataset_id,
                offset=i._page_offset or 0,
                limit=min(i._page_size, i._limit),
            )
            query["queries"].append(_params)

        return query

    def execute(self, **params):
        _params = self._build_query()
        _params.update(**params)
        response = self._client.post("/v2/batch_query", _params)
        return response


class QueryFile(QueryBase):
    """
    A QueryFile API request wrapper that generates a request for an object content query,
    and can iterate through streaming result sets.
    """

    # The default number of records fetched in one go.
    DEFAULT_PAGE_SIZE = 1000

    def __init__(
        self,
        file_id,
        fields=None,
        exclude_fields=None,
        filters=None,
        limit=float("inf"),
        page_size=DEFAULT_PAGE_SIZE,
        output_format="json",
        header=True,
        **kwargs,
    ):
        """
        Creates a new QueryFile object.

        :Parameters:
          - `file_id`: Unique ID of file to query.
          - `fields` (optional): List of specific fields to retrieve.
          - `exclude_fields` (optional): List of specific fields to exclude.
          - `filters` (optional): Filter or List of filter objects.
          - `limit` (optional): Maximum number of query results to return.
          - `page_size` (optional): Number of results to fetch per query page.
          - `output_format` (optional): Format of query results (json, csv or tsv)
          - `header` (optional): Returns header in response if output_format is 'csv' or 'tsv'
        """
        self._file_id = file_id
        self._data_url = "/v2/objects/{0}/data".format(file_id)
        self._fields_url = "/v2/objects/{0}/fields".format(file_id)
        self._fields = fields
        self._exclude_fields = exclude_fields
        self._output_format = output_format
        self._header = header
        self._error = None

        if filters:
            if isinstance(filters, Filter):
                filters = [filters]
        else:
            filters = []
        self._filters = filters

        # init response and cursor
        self._response = None
        # Limit defines the total number of results that will be returned
        # from a query involving 1 or more pagination requests.
        self._limit = limit
        # Page offset can only be set by execute(). It is always set to the
        # current absolute offset contained in the buffer.
        self._page_size = int(page_size)
        # Page offset can only be set by execute(). It is always set to the
        # current absolute offset contained in the buffer.
        self._page_offset = None
        # slice is set when the QueryFile is being sliced.
        # In this case, __iter__() and next() will not
        # reset the page_offset to 0 before iterating.
        self._slice = None

        # parameter error checking
        if self._limit < 0:
            raise Exception("'limit' parameter must be >= 0")

        if not 0 < self._page_size <= self.MAX_PAGE_SIZE:
            raise Exception(
                "'page_size' parameter must be in " "range [1, {}]".format(
                    self.MAX_PAGE_SIZE
                )
            )

        # Set up the QuartzBioClient
        # (kwargs overrides pre-set, which overrides global)
        self._client = kwargs.get("client") or self._client or client

    def _clone(self, filters=None, limit=None):
        new = self.__class__(
            self._file_id,
            limit=self._limit,
            fields=self._fields,
            exclude_fields=self._exclude_fields,
            page_size=self._page_size,
            output_format=self._output_format,
            header=self._header,
            client=self._client,
        )

        new._filters += self._filters

        if filters:
            new._filters += filters

        if limit is not None:
            new._limit = limit

        return new

    def _build_query(self, **kwargs):
        q = {"output_format": self._output_format}

        if self._filters:
            filters = self._process_filters(self._filters)
            if len(filters) > 1:
                q["filters"] = [{"and": filters}]
            else:
                q["filters"] = filters

        if self._fields is not None:
            q["fields"] = self._fields

        if self._exclude_fields is not None:
            q["exclude_fields"] = self._exclude_fields

        # Add or modify query parameters
        # (used by BatchQuery and facets)
        q.update(**kwargs)

        return q

    def execute(self, offset=0, **query):
        """
        Executes a query. Additional query parameters can be passed
        as keyword arguments.

        Returns: The request parameters and the raw query response.
        """
        _params = self._build_query(**query)
        self._page_offset = offset

        _params.update(
            offset=self._page_offset, limit=min(self._page_size, self._limit)
        )

        logger.debug(
            "executing query. from/limit: %6d/%d"
            % (_params["offset"], _params["limit"])
        )

        # If the request results in a QuartzBioError (ie bad filter) set the error.
        try:
            self._response = self._client.post(self._data_url, _params)

            if (
                getattr(self, "_header", None)
                and self._output_format in ("csv", "tsv")
                and not getattr(self, "_header_fields", None)
            ):
                self._header_fields = self.fields()

                separator_mappings = {"csv": ",", "tsv": "\t"}
                sep = separator_mappings[self._output_format]

                self._response["results"].insert(0, sep.join(self._header_fields))

        except QuartzBioError as e:
            self._error = e
            raise

        logger.debug(
            "query response took: {} ms, total: {}".format(
                self._response["took"], self._response["total"]
            )
        )
        return _params, self._response

    def fields(self):
        """Returns all expected fields that will be found in the results."""

        fields = [f for f in self._client.get(self._fields_url, {})["fields"]]
        if self._fields:
            fields = [f for f in fields if f in self._fields]
        if self._exclude_fields:
            fields = [f for f in fields if f not in self._exclude_fields]

        return fields

    def __len__(self):
        base_len = super(QueryFile, self).__len__()
        if base_len == float("inf"):
            raise TypeError(
                "Unable to determine the number of records for the "
                "current file-based query"
            )
        return base_len



import unittest
import dash
import dash_html_components as html
import time
from http import cookies
from unittest import mock

from quartzbio.contrib.dash import QuartzBioAuth

from .credentials import OAUTH_CLIENT_ID
from .credentials import OAUTH_TOKEN


endpoints = {
    "protected": {
        "get": [
            "/_dash-layout",
            "/_dash-routes",
            "/_dash-dependencies",
            "/_dash-component-suites/dash_html_components/bundle.js",
        ],
        "post": ["/_dash-update-component"],
    },
    "unprotected": {"get": ["/"], "post": []},
}


def get_cookie(res, cookie_name):
    headers = res.headers.to_list()
    set_cookie_strings = [
        h for h in headers if (h[0] == "Set-Cookie" and cookie_name in h[1])
    ]
    try:
        cookie_string = set_cookie_strings[0][1]
    except IndexError as e:
        print(cookie_name)
        for header in headers:
            print(header)
        print(set_cookie_strings)
        raise e

    cookie = cookies.SimpleCookie(cookie_string)
    access_granted_cookie = cookie[list(cookie.keys())[0]].value
    return access_granted_cookie


def create_apps(client_id):
    app_permissions = ["private"]
    apps = dict((k, dash.Dash(k)) for k in app_permissions)
    for app in list(apps.values()):
        app.scripts.config.serve_locally = True
    auths = dict(
        (
            k,
            QuartzBioAuth(
                apps[k],
                "http://localhost:5000",
                client_id,
                secret_key="test-secret-key",
                salt="test-salt",
            ),
        )
        for k in app_permissions
    )
    apps["unregistered"] = dash.Dash("unregistered")
    apps["unregistered"].scripts.config.serve_locally = True
    return apps, auths


class ProtectedViewsTest(unittest.TestCase):
    def setUp(self):
        self._oauth_client_id = OAUTH_CLIENT_ID
        self._oauth_token = OAUTH_TOKEN
        self.longMessage = True

    def test_protecting_all_views(self):
        apps = create_apps(self._oauth_client_id)[0]
        self.assertEqual(
            (
                len(endpoints["protected"]["get"])
                + len(endpoints["unprotected"]["get"])
                + len(endpoints["protected"]["post"])
                + len(endpoints["unprotected"]["post"])
            ),
            # Subtract 1 for /<path> -> /<path:path>
            len([k for k in apps["unregistered"].server.url_map.iter_rules()]) - 1,
        )

    def test_unauthenticated_view(self):
        apps = create_apps(self._oauth_client_id)[0]
        for app_name in ["unregistered"]:
            app = apps[app_name]
            app.layout = html.Div()
            client = app.server.test_client()
            for endpoint in (
                endpoints["protected"]["get"] + endpoints["unprotected"]["get"]
            ):
                res = client.get(endpoint)
                test_name = "{} at {} ({})".format(res.status_code, endpoint, app_name)

                self.assertEqual(res.status_code, 200, test_name)

    def test_403_on_protected_endpoints_without_cookie(self):
        apps = create_apps(self._oauth_client_id)[0]
        for app in [apps["private"]]:
            app.layout = html.Div()
            client = app.server.test_client()
            for endpoint in endpoints["protected"]["get"]:
                res = client.get(endpoint)
                self.assertEqual(res.status_code, 403, endpoint)

            for endpoint in endpoints["unprotected"]["get"]:
                res = client.get(endpoint)
                self.assertEqual(res.status_code, 200, endpoint)

            # TODO - check 200 on post of unprotected endpoints?
            for endpoint in endpoints["protected"]["post"]:
                res = client.post(endpoint)
                self.assertEqual(res.status_code, 403, endpoint)

    def check_endpoints(self, auth, app, oauth_token, cookies=tuple(), all_200=False):
        def get_client():
            client = app.server.test_client()
            client.set_cookie("/", "quartzbio_oauth_token", oauth_token)
            for cookie in cookies:
                client.set_cookie("/", cookie["name"], cookie["value"])
            return client

        for endpoint in endpoints["unprotected"]["get"] + endpoints["protected"]["get"]:
            client = get_client()  # use a fresh client for every endpoint
            res = client.get(endpoint)
            test_name = "{} at {} as {}".format(res.status_code, endpoint, oauth_token)
            self.assertEqual(res.status_code, 200, test_name)
        return res

    def test_protected_endpoints_with_auth_cookie(self):
        apps, auths = create_apps(self._oauth_client_id)

        for app_name, app in apps.items():
            if app_name != "unregistered":
                app.layout = html.Div()
                self.check_endpoints(auths[app_name], app, self._oauth_token)

    def test_auth_cookie_caches_calls_to_quartzbio(self):
        app = dash.Dash()
        app.scripts.config.serve_locally = True
        auth = QuartzBioAuth(
            app,
            "https://localhost:5000",
            self._oauth_client_id,
            secret_key="test-secret-key",
            salt="test-salt",
        )
        app.layout = html.Div()

        creator = self._oauth_token
        f = "quartzbio.contrib.dash.quartzbio_auth.QuartzBioAuth.check_view_access"  # noqa
        with mock.patch(f, wraps=auth.check_view_access) as wrapped:
            self.check_endpoints(auth, app, creator)
            res = self.check_endpoints(auth, app, creator)

            n_endpoints = len(endpoints["protected"]["get"]) + len(
                endpoints["unprotected"]["get"]
            )

            self.assertEqual(wrapped.call_count, n_endpoints * 2)

            access_granted_cookie = get_cookie(res, auth.AUTH_COOKIE_NAME)
            self.check_endpoints(
                auth,
                app,
                creator,
                ({"name": auth.AUTH_COOKIE_NAME, "value": access_granted_cookie},),
            )
            self.assertEqual(wrapped.call_count, n_endpoints * 2)

            # Regenerate tokens with a shorter expiration
            # User's won't actually do this in practice, we're
            # just doing it to shorten up the expiration from 5 min
            # to 10 seconds
            auth.config["permissions_cache_expiry"] = 10
            auth.create_access_codes()
            res = self.check_endpoints(auth, app, creator)
            self.assertEqual(wrapped.call_count, n_endpoints * 3)

            # Using the same auth cookie should prevent an
            # additional access call
            access_granted_cookie = get_cookie(res, auth.AUTH_COOKIE_NAME)
            self.check_endpoints(
                auth,
                app,
                creator,
                ({"name": auth.AUTH_COOKIE_NAME, "value": access_granted_cookie},),
            )
            self.assertEqual(wrapped.call_count, (n_endpoints * 3))

            # But after the expiration time (10 seconds), another call to
            # QuartzBio should be made
            time.sleep(10)
            self.check_endpoints(auth, app, creator)
            self.assertEqual(wrapped.call_count, (n_endpoints * 4))


class LoginFlow(unittest.TestCase):
    def login_success(self):
        oauth_client_id = OAUTH_CLIENT_ID
        oauth_token = OAUTH_TOKEN

        app = dash.Dash()
        app.config.scripts.serve_locally = True
        QuartzBioAuth(
            app, "https://quartzbio-dash-auth-app.herokuapp.com", oauth_client_id
        )
        app.layout = html.Div()
        client = app.server.test_client()
        csrf_token = get_cookie(client.get("/"), "_csrf_token")
        client.set_cookie("/", "_csrf_token", csrf_token)

        res = client.post(
            "_login",
            headers={
                "Authorization": "Bearer {}".format(oauth_token),
                "X-CSRFToken": csrf_token,
            },
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(get_cookie(res, "quartzbio_oauth_token"), oauth_token)




import sys
import os
import json
import tempfile

import mock

from .helper import QuartzBioTestCase

from quartzbio.cli import main
from quartzbio import DatasetTemplate
from quartzbio import Vault
from quartzbio.resource.object import Object
from quartzbio.errors import NotFoundError
from quartzbio.cli.data import should_exclude
from quartzbio.test.client_mocks import fake_vault_all
from quartzbio.test.client_mocks import fake_vault_create
from quartzbio.test.client_mocks import fake_object_all
from quartzbio.test.client_mocks import fake_object_create
from quartzbio.test.client_mocks import fake_object_retrieve
from quartzbio.test.client_mocks import fake_dataset_tmpl_create
from quartzbio.test.client_mocks import fake_dataset_tmpl_retrieve
from quartzbio.test.client_mocks import fake_dataset_import_create
from quartzbio.test.client_mocks import fake_dataset_get_or_create


def raise_not_found(*args, **kwargs):
    raise NotFoundError


def upload_path(*args, **kwargs):
    return "/"


class CLITests(QuartzBioTestCase):
    def setUp(self):
        super(CLITests, self).setUp()


class CreateDatasetTests(CLITests):
    @mock.patch("quartzbio.resource.Vault.all")
    @mock.patch("quartzbio.resource.Object.all")
    @mock.patch("quartzbio.resource.Dataset.get_or_create_by_full_path")
    @mock.patch("quartzbio.resource.DatasetTemplate.create")
    @mock.patch("quartzbio.resource.DatasetTemplate.retrieve")
    def _test_create_dataset(
        self, args, TmplRetrieve, TmplCreate, DatasetCreate, ObjectAll, VaultAll
    ):
        TmplRetrieve.side_effect = fake_dataset_tmpl_retrieve
        TmplCreate.side_effect = fake_dataset_tmpl_create
        DatasetCreate.side_effect = fake_dataset_get_or_create
        ObjectAll.side_effect = fake_object_all
        VaultAll.side_effect = fake_vault_all
        return main.main(args)

    def test_create_dataset(self):
        args = [
            "create-dataset",
            "quartzbio:test_vault:/test-dataset",
            "--capacity",
            "small",
        ]
        ds = self._test_create_dataset(args)
        self.assertEqual(ds.filename, "test-dataset")
        self.assertEqual(ds.path, "/test-dataset")

    def test_create_dataset_by_filename(self):
        args = [
            "create-dataset",
            "quartzbio:test_vault:/test-dataset-filename",
            "--capacity",
            "small",
            "--tag",
            "tag_test",
            "--metadata",
            "TEST=tag",
            "--metadata",
            "TEST2=tag2",
        ]
        ds = self._test_create_dataset(args)
        self.assertEqual(ds.filename, "test-dataset-filename")
        self.assertEqual(ds.path, "/test-dataset-filename")
        self.assertEqual(ds.capacity, "small")
        self.assertEqual(ds.tags, ["tag_test"])
        self.assertEqual(ds.metadata, dict(TEST="tag", TEST2="tag2"))

    def _validate_tmpl_fields(self, fields):
        for f in fields:
            if f.name == "name":
                self.assertEqual(f.entity_type, "gene")
            elif f.name == "variants":
                self.assertEqual(f.entity_type, "variant")
                self.assertEqual(f.is_list, True)
                self.assertEqual(f.data_type, "auto")
            elif f.name == "aliases":
                self.assertEqual(f.data_type, "string")

    def test_create_dataset_upload_template(self):
        template_path = os.path.join(os.path.dirname(__file__), "data/template.json")
        args = [
            "create-dataset",
            "quartzbio:test_vault:/test-dataset",
            "--template-file",
            template_path,
            "--capacity",
            "medium",
        ]  # noqa

        ds = self._test_create_dataset(args)
        self.assertEqual(ds.description, "Created with dataset template: 100")

    @mock.patch("quartzbio.resource.DatasetTemplate.create")
    def test_create_dataset_template_id(self, TmplCreate):
        # create template
        template_path = os.path.join(os.path.dirname(__file__), "data/template.json")
        with open(template_path, "r") as fp:
            tpl_json = json.load(fp)

        TmplCreate.side_effect = fake_dataset_tmpl_create
        tpl = DatasetTemplate.create(**tpl_json)
        args = [
            "create-dataset",
            "quartzbio:test_vault:/test-dataset",
            "--template-id",
            str(tpl.id),
            "--capacity",
            "small",
        ]  # noqa

        ds = self._test_create_dataset(args)
        self.assertEqual(
            ds.description, "Created with dataset template: {0}".format(tpl.id)
        )


class ImportTests(CLITests):
    @mock.patch("quartzbio.resource.DatasetTemplate.retrieve")
    @mock.patch("quartzbio.resource.DatasetTemplate.create")
    @mock.patch("quartzbio.resource.Vault.get_or_create_uploads_path")
    @mock.patch("quartzbio.resource.Vault.get_by_full_path")
    @mock.patch("quartzbio.resource.Vault.all")
    @mock.patch("quartzbio.resource.Object.all")
    @mock.patch("quartzbio.resource.Object.upload_file")
    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.DatasetImport.create")
    @mock.patch("quartzbio.resource.Dataset.get_or_create_by_full_path")
    def _test_import_file(
        self,
        args,
        DatasetGetCreate,
        DatasetImportCreate,
        ObjectCreate,
        UploadFile,
        ObjectAll,
        VaultAll,
        VaultLookup,
        UploadPath,
        TmplCreate,
        TmplRetrieve,
    ):
        DatasetGetCreate.side_effect = fake_dataset_get_or_create
        DatasetImportCreate.side_effect = fake_dataset_import_create
        ObjectCreate.side_effect = fake_object_create
        UploadFile.side_effect = fake_object_create
        ObjectAll.side_effect = fake_object_all
        VaultAll.side_effect = fake_vault_all
        UploadPath.side_effect = upload_path
        VaultLookup.side_effect = fake_vault_create
        TmplRetrieve.side_effect = fake_dataset_tmpl_retrieve
        TmplCreate.side_effect = fake_dataset_tmpl_create

        # returns (imports_list, dataset)
        return main.main(args)

    def test_import_file(self):
        _, file_ = tempfile.mkstemp(suffix=".txt")
        with open(file_, "w") as fp:
            fp.write("blargh")

        args = [
            "import",
            "--create-dataset",
            "--tag",
            "hello",
            "--follow",
            "--commit-mode",
            "overwrite",
            "quartzbio:mock_vault:/test-dataset",
            file_,
        ]

        imports, ds = self._test_import_file(args)
        self.assertEqual(ds.full_path, "quartzbio:mock_vault:/test-dataset")
        self.assertEqual(ds.tags, ["hello"])

        # should be a manifest with a single file
        self.assertEqual(len(imports[0].manifest["files"]), 1)
        self.assertEqual(imports[0]["commit_mode"], "overwrite")

        # Non-existent file
        args = [
            "import",
            "--create-dataset",
            "--tag",
            "hello",
            "--follow",
            "quartzbio:mock_vault:/test-dataset",
            "not/a/real/path/file.txt",
        ]

        with self.assertRaises(ValueError):
            self._test_import_file(args)

    def test_import_tilde(self):
        _, file_ = tempfile.mkstemp(suffix=".txt")
        with open(file_, "w") as fp:
            fp.write("blargh")

        for dataset_path in [
            "~/test-dataset",
            "~:/test-dataset",
        ]:
            args = ["import", "--create-dataset", dataset_path, file_]
            imports, ds = self._test_import_file(args)
            self.assertEqual(ds.filename, "test-dataset")
            # should be a manifest with a single file
            self.assertEqual(len(imports[0].manifest["files"]), 1)

    def test_import_remote_file(self):
        args = [
            "import",
            "--create-dataset",
            "--tag",
            "hello",
            "--remote-source",
            "quartzbio:mock_vault:/test-dataset",
            "/this/is/a/remote/file",
            "/this/remote/*/path",
        ]

        imports, ds = self._test_import_file(args)
        self.assertEqual(ds.full_path, "quartzbio:mock_vault:/test-dataset")
        self.assertEqual(ds.tags, ["hello"])
        # should be two imports
        self.assertEqual(len(imports), 2)
        for import_ in imports:
            self.assertEqual(import_["object_id"], 100)
            self.assertTrue("manifest" not in import_)

    def test_import_file_template(self):
        template_path = os.path.join(os.path.dirname(__file__), "data/template.json")

        args = [
            "import",
            "--create-dataset",
            "--tag",
            "hello",
            "--template-file",
            template_path,
            "--remote-source",
            "quartzbio:mock_vault:/test-dataset",
            "/this/is/a/remote/file",
        ]

        imports, ds = self._test_import_file(args)
        self.assertEqual(ds.full_path, "quartzbio:mock_vault:/test-dataset")
        self.assertEqual(ds.tags, ["hello"])
        self.assertEqual(len(imports), 1)
        import_ = imports[0]
        self.assertEqual(import_["object_id"], 100)
        self.assertTrue("manifest" not in import_)

        with open(template_path, "r") as fp:
            template = json.load(fp)
        for key in [
            "reader_params",
            "annotator_params",
            "validation_params",
            "entity_params",
        ]:
            self.assertEqual(import_[key], template[key])


class UploadTests(CLITests):
    @mock.patch("quartzbio.resource.apiresource.ListableAPIResource._retrieve_helper")
    @mock.patch("quartzbio.resource.Vault.get_by_full_path")
    @mock.patch("quartzbio.resource.Vault.all")
    @mock.patch("quartzbio.resource.Object.all")
    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.Object.upload_file")
    @mock.patch("quartzbio.global_search.GlobalSearch.filter")
    def _test_upload_command(
        self,
        args,
        GlobalSearch,
        ObjectUpload,
        ObjectCreate,
        ObjectAll,
        VaultAll,
        VaultLookup,
        RetrieveHelper,
        **kwargs,
    ):
        GlobalSearch.side_effect = fake_object_all
        ObjectUpload.side_effect = fake_object_create
        ObjectAll.side_effect = fake_object_all
        ObjectCreate.side_effect = fake_object_create
        VaultAll.side_effect = fake_vault_all
        VaultLookup.side_effect = fake_vault_create

        if "fail_lookup" in kwargs:
            RetrieveHelper.side_effect = raise_not_found
        else:
            RetrieveHelper.side_effect = fake_object_retrieve

        main.main(args)

    def test_upload_file(self):
        _, file_ = tempfile.mkstemp(suffix=".txt")
        with open(file_, "w") as fp:
            fp.write("blargh")

        args = ["upload", "--full-path", "quartzbio:test_vault:/test-folder", file_]
        with self.assertRaises(NotFoundError):
            self._test_upload_command(args, fail_lookup=True)

        # pass -p to create destination
        args = [
            "upload",
            "--full-path",
            "quartzbio:test_vault:/test-folder",
            "--create-full-path",
            file_,
        ]
        self._test_upload_command(args)

    def test_upload_directories(self):
        folder_ = tempfile.mkdtemp(suffix=".txt")
        inner_folder_ = tempfile.mkdtemp(suffix=".txt", dir=folder_)
        _, file_ = tempfile.mkstemp(suffix=".txt", dir=inner_folder_)
        with open(file_, "w") as fp:
            fp.write("blargh")

        args = [
            "upload",
            "--full-path",
            "quartzbio:test_vault:/test-folder-upload",
            folder_,
        ]
        with self.assertRaises(NotFoundError):
            self._test_upload_command(args, fail_lookup=True)

        # pass -p to create destination
        args = [
            "upload",
            "--full-path",
            "quartzbio:test_vault:/test-folder-upload",
            "--create-full-path",
            folder_,
        ]
        self._test_upload_command(args)

        # Test multiprocess
        args = [
            "upload",
            "--num-processes",
            "2",
            "--full-path",
            "quartzbio:test_vault:/test-folder-upload",
            folder_,
        ]
        with self.assertRaises(NotFoundError):
            self._test_upload_command(args, fail_lookup=True)

    @mock.patch("quartzbio.resource.apiresource.ListableAPIResource._retrieve_helper")
    @mock.patch("quartzbio.resource.Object.all")
    @mock.patch("quartzbio.resource.Object.create")
    @mock.patch("quartzbio.resource.Vault.create")
    def test_create_folder(self, VaultCreate, ObjectCreate, ObjectAll, RetrieveHelper):
        VaultCreate.side_effect = fake_vault_create
        ObjectAll.side_effect = fake_object_all
        ObjectCreate.side_effect = fake_object_create

        vault = Vault.create("my-vault")

        # create folder

        RetrieveHelper.side_effect = fake_object_retrieve
        full_path = vault.full_path + ":/new_folder"
        f = Object.create_folder(vault, full_path)
        self.assertEqual(f.full_path, full_path)

    def test_should_exclude(self):
        exclude = ["~/test", "~/test2"]
        self.assertFalse(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/test*"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["*file.txt"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["*file.json"]
        self.assertFalse(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/test3/"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/test3"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/*"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/test3/*"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["file.txt"]
        self.assertFalse(should_exclude("~/test3/file.txt", exclude))

        exclude = ["~/test3/file.txt"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["*file.txt"]
        self.assertTrue(should_exclude("~/test3/file.txt", exclude))

        exclude = ["*/folder/*/file.txt"]
        self.assertTrue(should_exclude("~/folder/2019-01-01/file.txt", exclude))

        exclude = ["*/folder/*/new/file.txt"]
        self.assertFalse(should_exclude("~/folder/2019-01-01/file.txt", exclude))

        exclude = ["folder"]
        self.assertFalse(should_exclude("~/folder/2019-01-01/file.txt", exclude))

        exclude = ["*folder"]
        self.assertTrue(should_exclude("~/folder/2019-01-01/file.txt", exclude))

        exclude = ["*folder*"]
        self.assertTrue(should_exclude("~/folder/2019-01-01/file.txt", exclude))


class DownloadTests(CLITests):
    @mock.patch("quartzbio.resource.Object.all")
    @mock.patch("quartzbio.resource.apiresource.DownloadableAPIResource.download")
    def _test_download_file(self, args, Download, ObjectAll, download_success=True):
        ObjectAll.side_effect = fake_object_all

        if download_success:
            Download.side_effect = lambda x: True
        else:
            Download.side_effect = Exception("Mock Download Fail")

        # returns (imports_list, dataset)
        return main.main(args)

    def test_download_file(self):
        args = ["download", "quartzbio:mock_vault:/test-file", "."]
        self._test_download_file(args)
        self.assertFalse(os.path.exists(".test-file"))

        args = ["download", "quartzbio:mock_vault:/test-file/*", "."]
        self._test_download_file(args)
        self.assertFalse(os.path.exists(".test-file"))

        # args needed
        args = ["download"]
        with self.assertRaises(SystemExit):
            self._test_download_file(args)

        # full path required
        args = ["download", "mypath"]
        with self.assertRaises(SystemExit):
            self._test_download_file(args)

        # local path required
        args = ["download", "my-vault:/mypath"]
        with self.assertRaises(SystemExit):
            self._test_download_file(args)

        args = ["download", "quartzbio:mock_vault:/test-file/*", "."]
        with self.assertRaises(Exception):
            self._test_download_file(args, download_success=False)

    @mock.patch("quartzbio.global_search.GlobalSearch.filter")
    @mock.patch("quartzbio.resource.apiresource.DownloadableAPIResource.download")
    @mock.patch("os.makedirs")
    @mock.patch("quartzbio.resource.object.Object.retrieve")
    @mock.patch("os.walk")
    @mock.patch("os.rmdir")
    @mock.patch("os.remove")
    @mock.patch("os.listdir")
    def _test_download_folder(
        self,
        args,
        ListDir,
        Remove,
        RmDir,
        Walk,
        ObjectRetrieve,
        MakeDirs,
        Download,
        GlobalSearch,
        download_success=True,
        expected_downloads=2,
        expected_delete_files=0,
        expected_delete_folders=0,
    ):
        # Create GlobalSearch return objects
        remote_objects = [
            fake_object_create(id=1, filename="test-folder", object_type="folder"),
            fake_object_create(
                id=2, filename="test-folder/subfolder", object_type="folder"
            ),
            fake_object_create(
                id=3, filename="test-folder/subfolder/file.txt", object_type="file"
            ),
            fake_object_create(
                id=4, filename="test-folder/subfolder/file2.csv", object_type="file"
            ),
            fake_vault_create(
                id=5,
                full_path="test-folder/subfolder/vault",
                name="vault",
                class_name="Vault",
                path="test-folder/subfolder/vault",
            ),
        ]
        GlobalSearch.return_value = remote_objects

        # Mock query to API to get md5
        def object_retrieve_side_effect(file_id):
            return [x for x in remote_objects if x.id == file_id][0]

        ObjectRetrieve.side_effect = object_retrieve_side_effect

        # Fail on demand
        if download_success:
            Download.side_effect = lambda x: True
        else:
            Download.side_effect = Exception("Mock Download Fail")

        # Create files that would be deleted by --delete
        Walk.return_value = [("/sample/root/dir", ["empty_folder"], ["local_file"])]
        ListDir.return_value = []

        # returns (imports_list, dataset)
        main.main(args)

        # Only downloads 'file.txt', folders are ignored
        self.assertEqual(Download.call_count, expected_downloads)
        self.assertEqual(Remove.call_count, expected_delete_files)
        self.assertEqual(RmDir.call_count, expected_delete_folders)

    def test_download_folder(self):
        args = ["download", "--recursive", "quartzbio:mock_vault:/test-folder", "."]
        self._test_download_folder(args)
        self.assertFalse(os.path.exists("./test-folder"))

        # Exclude all files
        args = [
            "download",
            "--recursive",
            "--exclude",
            "*",
            "quartzbio:mock_vault:/test-folder/",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=0)

        # ignore vaults
        args = [
            "download",
            "--recursive",
            "quartzbio:mock_vault:/test-folder/",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=2)

        # Exclude suffix
        args = [
            "download",
            "--recursive",
            "--exclude",
            "*txt",
            "quartzbio:mock_vault:/test-folder/",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=1)

        # exclude folder name
        args = [
            "download",
            "--recursive",
            "--exclude",
            "*subfolder*",
            "quartzbio:mock_vault:/test-folder/",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=0)

        # Exclude all files, then includes txt files
        args = [
            "download",
            "--recursive",
            "--exclude",
            "*",
            "--include",
            "*.txt",
            "quartzbio:mock_vault:/test-folder/",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=1)

        # Test dry run
        args = [
            "download",
            "--recursive",
            "--dry-run",
            "quartzbio:mock_vault:/test-folder",
            ".",
        ]
        self._test_download_folder(args, expected_downloads=0)

        # Test delete
        args = [
            "download",
            "--recursive",
            "--delete",
            "quartzbio:mock_vault:/test-folder",
            ".",
        ]
        self._test_download_folder(
            args, expected_delete_files=1, expected_delete_folders=1
        )

        # args needed
        args = ["download", "--recursive"]
        with self.assertRaises(SystemExit):
            self._test_download_folder(args)

        # full path required
        args = ["download", "--recursive", "mypath"]
        with self.assertRaises(SystemExit):
            self._test_download_folder(args)

        # local path required
        args = ["download", "--recursive", "my-vault:/mypath"]
        with self.assertRaises(SystemExit):
            self._test_download_folder(args)

        args = ["download", "--recursive", "quartzbio:mock_vault:/test-file/*", "."]
        with self.assertRaises(Exception):
            self._test_download_folder(args, download_success=False)


class LsTests(CLITests):
    @mock.patch("quartzbio.resource.object.Object.all")
    def _test_ls(self, args, ObjectAll, fake_objects=[]):
        ObjectAll.return_value = fake_objects

        # Mocking print function is different in
        # python 2 and 3
        if sys.version_info.major == 3:
            builtin_module_name = "builtins.print"
        else:
            builtin_module_name = "__builtin__.print"
        with mock.patch(builtin_module_name) as Print:
            main.main(args)

        main.main(args)
        return Print

    def test_download_folder(self):
        remote_objects = [
            fake_object_create(
                last_modified="2020", filename="folder", object_type="folder"
            ),
            fake_object_create(
                last_modified="2021", filename="old-file", object_type="file"
            ),
            fake_object_create(
                last_modified="2022", filename="new-file", object_type="file"
            ),
        ]
        args = ["ls", "~"]
        mock_print = self._test_ls(args, fake_objects=remote_objects)
        self.assertEqual(mock_print.call_count, len(remote_objects))

        mock_print = self._test_ls(args, fake_objects=[])
        self.assertEqual(mock_print.call_count, 1)
        self.assertIn("No file(s) found", mock_print.call_args[0][0])


class QueueTests(CLITests):
    def test_show_queue(self):
        """Simple test to print the queue"""
        main.main(["queue"])



from .apiresource import ListableAPIResource
from .apiresource import DeletableAPIResource
from .apiresource import CreateableAPIResource
from .apiresource import DownloadableAPIResource
from .task import Task
from .quartzbio_object import convert_to_quartzbio_object

import time


class DatasetExport(
    CreateableAPIResource,
    ListableAPIResource,
    DownloadableAPIResource,
    DeletableAPIResource,
):
    """
    DatasetExport represent an export task that takes
    a Dataset or filtered Dataset (Query) and exports
    the contents to a flat file (CSV, JSON, or XLSX).

    For interactive use, DatasetExport can be "followed" to watch
    the progression of the task.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("documents_count", "Records"),
        ("format", "Format"),
        ("status", "Status"),
        ("created_at", "Created"),
    )

    @property
    def dataset(self):
        response = self._client.get(self["dataset"]["url"], {})
        return convert_to_quartzbio_object(response, client=self._client)

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        if self.status == "queued":
            print("Waiting for export (id = {0}) to start...".format(self.id))

        export_status = self.status
        while self.status in ["queued", "running"]:
            if self.status != export_status:
                print("Export is now {0} (was {1})".format(self.status, export_status))
                export_status = self.status

            if self.status == "running":
                print(
                    "Export '{0}' is {1}: {2}/{3} records exported".format(
                        self.id,
                        self.status,
                        self.metadata.get("progress", {}).get("processed_records", 0),  # noqa
                        self.documents_count,
                    )
                )

            if not loop:
                return

            time.sleep(sleep_seconds)
            self.refresh()

        if self.status == "completed":
            print(
                "Export complete! Run <export_obj>"
                ".download(path=<some_path>) to download."
            )
